"""
This module contains the QC Lab package initialization.
"""

from qclab.simulation import Simulation
from qclab.data import Data
from qclab.model import Model
from qclab.algorithm import Algorithm
from qclab.constants import Constants
from qclab.utils import configure_memory_logger
import qclab.models
import qclab.algorithms
import qclab.dynamics
import qclab.interfaces

configure_memory_logger()

try:
    from importlib.metadata import version as _pkg_version
except Exception:
    from importlib_metadata import version as _pkg_version

try:
    __version__ = _pkg_version("qclab")
except Exception:
    __version__ = "0+unknown"



"""
This module runs the spin-boson model using the sparse and vectorization
decorators and compares the result against a reference calculation.
"""

import pytest
import numpy as np
import os
import time
from qclab import Simulation, Data
import qclab.ingredients as ingredients
from qclab.algorithms import MeanField
from qclab.functions import vectorize_ingredient, make_ingredient_sparse
from qclab.dynamics import serial_driver
from qclab.models import SpinBoson

try:
    from tests.reference_settings import model_sim_settings, model_settings
except ImportError:
    from reference_settings import model_sim_settings, model_settings


def test_vectorize_decorator():
    """
    This test checks the ``vectorize_ingredient`` decorator.

    It runs a spin-boson model simulation using the ``MeanField`` algorithm
    where the quantum-classical and classical Hamiltonians are vectorized using
    the decorator. The results are compared against a reference calculation.
    """
    print("Testing vectorization decorator")
    reference_folder = os.path.join(os.path.dirname(__file__), "reference/")
    sim = Simulation(model_sim_settings["SpinBoson"])
    sim.model = SpinBoson(model_settings["SpinBoson"])
    sim.model.initialize_constants()
    sim.algorithm = MeanField()

    @vectorize_ingredient
    def h_q(model, parameters, **kwargs):
        E = model.constants.get("E")
        V = model.constants.get("V")
        return np.array([[E, V], [V, -E]])

    @vectorize_ingredient
    def h_qc(model, parameters, **kwargs):
        z = kwargs["z"]
        A = model.constants.get("A")
        diag_coupling = model.constants.diagonal_linear_coupling
        out = np.zeros((2, 2), dtype=complex)
        for i in range(A):
            out[0, 0] += diag_coupling[0, i] * (z[i] + np.conj(z[i]))
            out[1, 1] += diag_coupling[1, i] * (z[i] + np.conj(z[i]))
        return out

    sim.model.ingredients.append(("h_q", h_q))
    sim.model.ingredients.append(("h_qc", h_qc))
    sim.initial_state["wf_db"] = np.zeros(
        sim.model.constants.num_quantum_states, dtype=complex
    )
    sim.initial_state["wf_db"][0] = 1j
    st = time.time()
    data = serial_driver(sim)
    et = time.time()
    print(f"Finished in {et - st:.2f} seconds.")
    data_correct = Data().load(os.path.join(reference_folder, "SpinBoson_MeanField.h5"))
    for key, val in data.data_dict.items():
        np.testing.assert_allclose(
            val, data_correct.data_dict[key], rtol=1e-5, atol=1e-8
        )


def test_make_sparse_decorator():
    """
    This test checks the ``make_ingredient_sparse`` decorator.

    It runs a spin-boson model simulation using the ``MeanField`` algorithm
    where the gradient of the quantum-classical Hamiltonian is made sparse using
    the decorator. The results are compared against a reference calculation.
    """
    print("Testing sparsity decorator")
    reference_folder = os.path.join(os.path.dirname(__file__), "reference/")
    sim = Simulation(model_sim_settings["SpinBoson"])
    sim.model = SpinBoson(model_settings["SpinBoson"])
    sim.model.initialize_constants()
    sim.algorithm = MeanField()

    @make_ingredient_sparse
    def dh_qc_dzc(model, parameters, **kwargs):
        inds, mels, shape = ingredients.dh_qc_dzc_diagonal_linear(
            model, parameters, **kwargs
        )
        out = np.zeros(shape, dtype=complex)
        np.add.at(out, inds, mels)
        return out

    sim.model.ingredients.append(("dh_qc_dzc", dh_qc_dzc))
    sim.initial_state["wf_db"] = np.zeros(
        sim.model.constants.num_quantum_states, dtype=complex
    )
    sim.initial_state["wf_db"][0] = 1j
    st = time.time()
    data = serial_driver(sim)
    et = time.time()
    print(f"Finished in {et - st:.2f} seconds.")
    data_correct = Data().load(os.path.join(reference_folder, "SpinBoson_MeanField.h5"))
    for key, val in data.data_dict.items():
        np.testing.assert_allclose(
            val, data_correct.data_dict[key], rtol=1e-5, atol=1e-8
        )


def test_sparse_and_vectorize_decorators():
    """
    This test checks the combination of the ``make_ingredient_sparse`` and ``vectorize_ingredient`` decorators.

    It runs a spin-boson model simulation using the ``MeanField`` algorithm
    where the gradient of the quantum-classical Hamiltonian is made sparse using
    the ``make_ingredient_sparse`` decorator and vectorized using the
    ``vectorize_ingredient`` decorator. The results are compared against a reference calculation.
    """
    print("Testing sparsity and vectorization decorators")
    reference_folder = os.path.join(os.path.dirname(__file__), "reference/")
    sim = Simulation(model_sim_settings["SpinBoson"])
    sim.model = SpinBoson(model_settings["SpinBoson"])
    sim.model.initialize_constants()
    sim.algorithm = MeanField()

    @make_ingredient_sparse
    @vectorize_ingredient
    def dh_qc_dzc(model, parameters, **kwargs):
        A = model.constants.get("A")
        diag_coupling = model.constants.diagonal_linear_coupling
        out = np.zeros((A, 2, 2), dtype=complex)
        for i in range(A):
            out[i, 0, 0] = diag_coupling[0, i]
            out[i, 1, 1] = diag_coupling[1, i]
        return out

    sim.model.ingredients.append(("dh_qc_dzc", dh_qc_dzc))
    sim.initial_state["wf_db"] = np.zeros(
        sim.model.constants.num_quantum_states, dtype=complex
    )
    sim.initial_state["wf_db"][0] = 1j
    st = time.time()
    data = serial_driver(sim)
    et = time.time()
    print(f"Finished in {et - st:.2f} seconds.")
    data_correct = Data().load(os.path.join(reference_folder, "SpinBoson_MeanField.h5"))
    for key, val in data.data_dict.items():
        np.testing.assert_allclose(
            val, data_correct.data_dict[key], rtol=1e-5, atol=1e-8
        )


if __name__ == "__main__":
    st = time.time()
    test_vectorize_decorator()
    et1 = time.time()
    print(f"Vectorization tests completed in {et1 - st:.2f} seconds.")
    test_make_sparse_decorator()
    et2 = time.time()
    print(f"Sparsity tests completed in {et2 - et1:.2f} seconds.")
    test_sparse_and_vectorize_decorators()
    et3 = time.time()
    print(f"Sparsity and vectorization tests completed in {et3 - et2:.2f} seconds.")
    print(f"All tests completed in {et3 - st:.2f} seconds.")



"""
This module contains the FewestSwitchesSurfaceHopping algorithm class.
"""

from functools import partial
from qclab.algorithm import Algorithm
from qclab import tasks


class FewestSwitchesSurfaceHopping(Algorithm):
    """
    Fewest switches surface hopping algorithm class.
    """

    def __init__(self, settings=None):
        if settings is None:
            settings = {}
        self.default_settings = {
            "fssh_deterministic": False,
            "gauge_fixing": "sign_overlap",
            "use_gauge_field_force": False,
        }
        super().__init__(self.default_settings, settings)

    initialization_recipe = [
        tasks.initialize_variable_objects,
        tasks.initialize_norm_factor,
        tasks.initialize_branch_seeds,
        tasks.initialize_z,
        tasks.update_h_q_tot,
        partial(
            tasks.diagonalize_matrix,
            matrix_name="h_q_tot",
            eigvals_name="eigvals",
            eigvecs_name="eigvecs",
        ),
        partial(
            tasks.update_eigvecs_gauge,
            gauge_fixing="phase_der_couple",
            eigvecs_previous_name="eigvecs",
        ),
        partial(tasks.copy_in_state, copy_name="eigvecs_previous", orig_name="eigvecs"),
        partial(
            tasks.update_vector_basis,
            input_vec_name="wf_db",
            basis_name="eigvecs",
            output_vec_name="wf_adb",
            adb_to_db=False,
        ),
        tasks.initialize_random_values_fssh,
        tasks.initialize_active_surface,
        tasks.initialize_dm_adb_0_fssh,
        tasks.update_act_surf_wf,
    ]

    update_recipe = [
        partial(tasks.copy_in_state, copy_name="eigvecs_previous", orig_name="eigvecs"),
        # Begin RK4 integration steps.
        tasks.update_classical_force,
        partial(
            tasks.update_quantum_classical_force,
            wf_db_name="act_surf_wf",
            wf_changed=True,
        ),
        tasks.update_z_rk4_k123,
        partial(tasks.update_classical_force, z_name="z_1"),
        partial(
            tasks.update_quantum_classical_force,
            wf_db_name="act_surf_wf",
            z_name="z_1",
            wf_changed=False,
        ),
        partial(tasks.update_z_rk4_k123, z_name="z", z_k_name="z_2", k_name="z_rk4_k2"),
        partial(tasks.update_classical_force, z_name="z_2"),
        partial(
            tasks.update_quantum_classical_force,
            wf_db_name="act_surf_wf",
            z_name="z_2",
            wf_changed=False,
        ),
        partial(
            tasks.update_z_rk4_k123,
            z_name="z",
            z_k_name="z_3",
            k_name="z_rk4_k3",
            dt_factor=1.0,
        ),
        partial(tasks.update_classical_force, z_name="z_3"),
        partial(
            tasks.update_quantum_classical_force,
            wf_db_name="act_surf_wf",
            z_name="z_3",
            wf_changed=False,
        ),
        tasks.update_z_rk4_k4,
        # End RK4 integration steps.
        tasks.update_wf_db_propagator,
        tasks.update_h_q_tot,
        partial(
            tasks.diagonalize_matrix,
            matrix_name="h_q_tot",
            eigvals_name="eigvals",
            eigvecs_name="eigvecs",
        ),
        tasks.update_eigvecs_gauge,
        partial(
            tasks.update_vector_basis,
            input_vec_name="wf_db",
            basis_name="eigvecs",
            output_vec_name="wf_adb",
            adb_to_db=False,
        ),
        tasks.update_hop_prob_fssh,
        tasks.update_hop_inds_fssh,
        tasks.update_hop_vals_fssh,
        tasks.update_z_hop,
        tasks.update_act_surf_hop,
        tasks.update_act_surf_wf,
    ]

    collect_recipe = [
        tasks.update_t,
        tasks.update_dm_db_fssh,
        tasks.update_quantum_energy_act_surf,
        tasks.update_classical_energy_fssh,
        tasks.collect_t,
        tasks.collect_dm_db,
        tasks.collect_quantum_energy,
        tasks.collect_classical_energy,
    ]


class FewestSwitchesSurfaceHoppingAbInitio(Algorithm):
    """
    Fewest switches surface hopping algorithm class implemented in the adiabatic basis
    for compatibility with ab initio calculations.
    """

    def __init__(self, settings=None):
        if settings is None:
            settings = {}
        self.default_settings = {
            "fssh_deterministic": False,
            "use_gauge_field_force": False,
            "update_wf_adb_eig_num_substeps": 10,
            "use_wf_overlaps_for_adb_connection": True,
        }
        super().__init__(self.default_settings, settings)

    initialization_recipe = [
        tasks.initialize_variable_objects,
        partial(tasks.copy_to_parameters, state_name="seed", parameters_name="seed"),
        tasks.initialize_norm_factor,
        tasks.initialize_branch_seeds,
        tasks.initialize_z,
        partial(
            tasks.update_ab_initio_property,
            property_dict={
                "energy": {"z": "z", "excited_amplitudes": True},
                "gradient": {"z": "z", "state_inds_gradient": None},
                "derivative_coupling": {
                    "z": "z",
                    "state_inds_derivative_coupling": None,
                },
            },
        ),
        tasks.update_h_q_tot,
        tasks.update_classical_force,
        tasks.update_derivative_coupling_dzc,
        partial(tasks.update_quantum_classical_force, wf_db_name="wf_adb"),
        partial(tasks.update_adb_connection, update_derivative_coupling=False),
        tasks.initialize_random_values_fssh,
        tasks.initialize_active_surface,
        tasks.initialize_dm_adb_0_fssh,
        partial(
            tasks.diagonalize_matrix,
            matrix_name="h_q_tot",
            eigvals_name="eigvals",
            eigvecs_name="eigvecs",
        ),
        tasks.update_act_surf_wf,
        tasks.update_quantum_energy_act_surf,
        tasks.update_classical_energy_fssh,
    ]

    update_recipe = [
        partial(
            tasks.copy_in_state,
            copy_name="aip_excited_amplitudes_previous",
            orig_name="aip_excited_amplitudes",
        ),
        partial(
            tasks.copy_in_state,
            copy_name="eigvecs_previous",
            orig_name="eigvecs",
        ),
        partial(
            tasks.copy_in_state,
            copy_name="adb_connection_previous",
            orig_name="adb_connection",
        ),
        partial(tasks.copy_in_state, copy_name="h_q_tot_previous", orig_name="h_q_tot"),
        partial(
            tasks.copy_in_state,
            copy_name="quantum_classical_force_previous",
            orig_name="quantum_classical_force",
        ),
        partial(
            tasks.copy_in_state,
            copy_name="classical_force_previous",
            orig_name="classical_force",
        ),
        partial(
            tasks.copy_in_state,
            copy_name="z_previous",
            orig_name="z",
        ),
        tasks.update_q_velocity_verlet,
        partial(
            tasks.update_ab_initio_property,
            property_dict={
                "energy": {"z": "z", "excited_amplitudes": True},
                "wf_overlaps": {
                    "z": "z",
                    "z_previous": "z_previous",
                    "amplitudes_previous": "aip_excited_amplitudes_previous",
                    "amplitudes_current": "aip_excited_amplitudes",
                },
            },
        ),
        # tasks.update_adb_connection,
        partial(tasks.update_adb_connection, update_derivative_coupling=True),
        tasks.update_h_q_tot,
        partial(
            tasks.diagonalize_matrix,
            matrix_name="h_q_tot",
            eigvals_name="eigvals",
            eigvecs_name="eigvecs",
        ),
        partial(
            tasks.update_wf_adb_hop_prob,
            update_hopping_probabilities=True,
        ),
        partial(
            tasks.update_hop_inds_fssh,
            hop_bool_name="hop_bool",
            hop_pairs_name="hop_pairs",
        ),
        partial(
            tasks.update_ab_initio_property,
            property_dict={
                "derivative_coupling": {
                    "calc_property": "hop_bool",
                    "z": "z",
                    "state_inds_derivative_coupling": "hop_pairs",
                },
            },
        ),
        tasks.update_derivative_coupling_dzc,
        partial(
            tasks.update_hop_vals_fssh,
            derivative_coupling_dzc_name="derivative_coupling_dzc",
        ),
        tasks.update_z_hop,
        tasks.update_act_surf_hop,
        tasks.update_act_surf_wf,
        partial(
            tasks.update_ab_initio_property,
            property_dict={
                "gradient": {"z": "z", "state_inds_gradient": "act_surf_ind"},
            },
        ),
        partial(tasks.update_quantum_classical_force, wf_db_name="act_surf_wf"),
        tasks.update_p_velocity_verlet,
        tasks.update_classical_force,
    ]

    collect_recipe = [
        tasks.update_t,
        tasks.update_dm_db_fssh,
        tasks.update_quantum_energy_act_surf,
        tasks.update_classical_energy_fssh,
        tasks.collect_t,
        partial(tasks.collect_dm_db, dm_db_name="dm_adb", dm_db_output_name="dm_adb"),
        tasks.collect_quantum_energy,
        tasks.collect_classical_energy,
    ]



"""
This module contains the parallel driver using the multiprocessing library.
"""

import multiprocessing
import logging
import copy
import numpy as np
import qclab.dynamics as dynamics
from qclab.utils import get_log_output, reset_log_output
from qclab import Data

logger = logging.getLogger(__name__)


def parallel_driver_multiprocessing(sim, seeds=None, data=None, num_tasks=None):
    """
    Parallel driver for the dynamics core using the python library multiprocessing.

    .. rubric:: Args
    sim: Simulation
        The Simulation object containing the model, algorithm, initial state, and settings.
    seeds: ndarray, optional
        An array of integer seeds for the trajectories. If None, seeds will be
        generated automatically.
    data: Data, optional
        A Data object for collecting output data. If None, a new Data object
        will be created.
    num_tasks: int, optional
        The number of tasks to use for parallel processing. If None, the
        number of available tasks will be used.

    .. rubric:: Returns
    data: Data
        The updated Data object containing collected output data.
    """
    # Clear any in-memory log output from previous runs.
    reset_log_output()
    # First initialize the model constants.
    sim.model.initialize_constants()
    if data is None:
        data = Data()
    if seeds is None:
        if len(data.data_dict["seed"]) > 0:
            offset = np.max(data.data_dict["seed"]) + 1
        else:
            offset = 0
        seeds = offset + np.arange(sim.settings.num_trajs, dtype=int)
        num_trajs = sim.settings.num_trajs
    else:
        num_trajs = len(seeds)
        logger.warning(
            "Setting sim.settings.num_trajs to the number of provided seeds: %s",
            num_trajs,
        )
        sim.settings.num_trajs = num_trajs
    if num_tasks is None:
        size = multiprocessing.cpu_count()
    else:
        size = num_tasks
    logger.info("Using %s tasks for parallel processing.", size)
    # Determine the number of batches required to execute the total number
    # of trajectories.
    if num_trajs % sim.settings.batch_size == 0:
        num_batches = num_trajs // sim.settings.batch_size
    else:
        num_batches = num_trajs // sim.settings.batch_size + 1
    logger.info(
        "Running %s batches with %s seeds in each batch.",
        num_batches,
        sim.settings.batch_size,
    )
    batch_seeds_list = (
        np.zeros((num_batches * sim.settings.batch_size), dtype=int) + np.nan
    )
    batch_seeds_list[:num_trajs] = seeds
    batch_seeds_list = batch_seeds_list.reshape((num_batches, sim.settings.batch_size))
    # Create the input data for each local simulation.
    sim.initialize_timesteps()
    local_input_data = [
        (
            copy.deepcopy(sim),
            {"seed": batch_seeds_list[n][~np.isnan(batch_seeds_list[n])].astype(int)},
            {},
            Data(batch_seeds_list[n][~np.isnan(batch_seeds_list[n])].astype(int)),
        )
        for n in range(num_batches)
    ]
    for i in range(num_batches):
        # Determine the batch size from the seeds in the State object.
        local_input_data[i][0].settings.batch_size = len(local_input_data[i][1]["seed"])
        logger.info(
            "Running batch %s with seeds %s.", i + 1, local_input_data[i][1]["seed"]
        )
    logger.info("Starting dynamics calculation.")
    with multiprocessing.Pool(processes=size) as pool:
        results = pool.starmap(dynamics.run_dynamics, local_input_data)
    logger.info("Dynamics calculation completed.")
    logger.info("Collecting results from all tasks.")
    for result in results:
        data.add_data(result)
    logger.info("Simulation complete.")
    # Attach collected log output.
    data.log = get_log_output()
    return data



"""
This module contains the Atomic Simulation Environment (ASE) Model class.
"""

import numpy as np
from qclab.functions import (
    vectorize_ingredient,
    make_ingredient_sparse,
    z_to_q,
    z_to_p,
    qp_to_z,
    dqdp_to_dzc,
)
from qclab.model import Model, Constants
from qclab import ingredients
from qclab.numerical_constants import (
    ANGSTROM_TO_BOHR,
)
import copy


class AbInitio(Model):
    """
    Model class for ab initio calculations.

    It is compatible with the ab initio algorithms implemented in QC Lab.
    """

    def __init__(self, constants=None):
        if constants is None:
            constants = {}
        self.default_constants = {
            "atom_positions": None,  # In units of Bohr.
            "atom_masses": None,  # In units of electron mass.
            "atom_names": None,
            "calculator_args": {},
            "num_quantum_states": None,
            "normal_mode": None,  # Unitless and normalilzed.
            "harmonic_frequency": None,  # In units of Hartrees.
            "energy_offset": 0,  # In units of Hartrees.
            "kBT": 0.00095,  # In units of Hartrees.
        }
        self.update_dh_qc_dzc = True
        self.update_h_q = False
        super().__init__(self.default_constants, constants)

    def _init_model(self, parameters, **kwargs):
        num_atoms = len(self.constants.atom_names)
        self.constants.num_classical_coordinates = num_atoms * 3
        self.constants.classical_coordinate_mass = (
            self.constants.atom_masses[np.newaxis] * np.ones((3, num_atoms))
        ).flatten()
        self.constants.classical_coordinate_weight = np.ones(
            self.constants.num_classical_coordinates
        )
        self.constants.init_position = self.constants.atom_positions.flatten()
        self.constants.finite_difference_delta = 1e-2

    def init_classical(self, parameters, **kwargs):
        # Temporarily set the masses to 1 for mass-weighted normal mode sampling.
        normal_modes = self.constants.normal_mode
        constants_original = copy.deepcopy(self.constants)
        self.constants = Constants()
        # Set number of coordinates to number of normal modes.
        self.constants.num_classical_coordinates = len(
            constants_original.harmonic_frequency
        )
        self.constants.classical_coordinate_mass = np.ones(
            self.constants.num_classical_coordinates
        )
        self.constants.harmonic_frequency = constants_original.harmonic_frequency
        self.constants.classical_coordinate_weight = self.constants.harmonic_frequency
        self.constants.kBT = constants_original.kBT
        z_normal_mode = ingredients.init_classical_wigner_harmonic(
            self, parameters, **kwargs
        )
        # Convert back to phase-space normal coordinates.
        q_normal_mode = z_to_q(
            z_normal_mode,
            self.constants.classical_coordinate_mass[np.newaxis],
            self.constants.classical_coordinate_weight[np.newaxis],
        )
        p_normal_mode = z_to_p(
            z_normal_mode,
            self.constants.classical_coordinate_mass[np.newaxis],
            self.constants.classical_coordinate_weight[np.newaxis],
        )
        # Convert back to Cartesian coordinates.
        q = (
            np.einsum("tm, cm ->tc", q_normal_mode, normal_modes)
            + constants_original.init_position
        )
        p = np.einsum("tm, cm ->tc", p_normal_mode, normal_modes)
        self.constants = constants_original
        z = qp_to_z(
            q,
            p,
            self.constants.classical_coordinate_mass[np.newaxis],
            self.constants.classical_coordinate_weight[np.newaxis],
        )
        return z

    def h_q(self, parameters, **kwargs):
        batch_size = kwargs["batch_size"]
        num_quantum_states = self.constants.num_quantum_states
        out = np.zeros(
            (batch_size, num_quantum_states, num_quantum_states), dtype=complex
        )
        return out

    @vectorize_ingredient
    def h_qc(self, parameters, **kwargs):
        z = kwargs["z"]
        num_quantum_states = self.constants.num_quantum_states
        traj_ind = kwargs["traj_ind"]
        has_ab_initio_property = "ab_initio_property" in parameters.keys()
        if has_ab_initio_property:
            properties = parameters["ab_initio_property"][traj_ind]
            if "energy" in properties.keys():
                diag_h_qc = (
                    properties["energy"][:num_quantum_states]
                    - self.constants.energy_offset
                )
                needs_energy = False
            else:
                needs_energy = True
        if not (has_ab_initio_property) or needs_energy:
            property_dict = {
                "energy": {"z": z[np.newaxis], "excited_amplitudes": True},
            }
            ab_initio_property_calculator, has_ab_intio_property_calculator = self.get(
                "ab_initio_property_calculator"
            )
            if has_ab_intio_property_calculator:
                properties = ab_initio_property_calculator(
                    self,
                    parameters,
                    batch_size=1,
                    property_dict=property_dict,
                )[0]
                diag_h_qc = (
                    properties["energy"][:num_quantum_states]
                    - self.constants.energy_offset
                )
            else:
                raise NameError("ab_initio_property_calculator must be provided")
        return np.diag(diag_h_qc)

    @make_ingredient_sparse
    @vectorize_ingredient
    def dh_qc_dzc(self, parameters, **kwargs):
        num_classical_coordinates = self.constants.num_classical_coordinates
        num_quantum_states = self.constants.num_quantum_states
        m = self.constants.classical_coordinate_mass
        h = self.constants.classical_coordinate_weight
        out = np.zeros(
            (num_classical_coordinates, num_quantum_states, num_quantum_states),
            dtype=complex,
        )
        z = kwargs["z"]
        traj_ind = kwargs["traj_ind"]
        has_ab_initio_property = "ab_initio_property" in parameters.keys()
        if has_ab_initio_property:
            properties = parameters["ab_initio_property"][traj_ind]
            if "gradient" in properties.keys():
                for state_ind in range(num_quantum_states):
                    # Convert to derivative w.r.t. zc.
                    out[:, state_ind, state_ind] = dqdp_to_dzc(
                        properties["gradient"][:, :, state_ind].flatten(),
                        None,
                        m,
                        h,
                    )
                needs_gradient = False
            else:
                needs_gradient = True
        if not (has_ab_initio_property) or needs_gradient:
            property_dict = {
                "gradient": {"z": z[np.newaxis], "state_inds_gradient": None},
            }
            ab_initio_property_calculator, has_ab_intio_property_calculator = self.get(
                "ab_initio_property_calculator"
            )
            if has_ab_intio_property_calculator:
                properties = ab_initio_property_calculator(
                    self,
                    parameters,
                    batch_size=1,
                    property_dict=property_dict,
                )[0]
                for state_ind in range(num_quantum_states):
                    # Convert to derivative w.r.t. zc.
                    out[:, state_ind, state_ind] = dqdp_to_dzc(
                        properties["gradient"][:, :, state_ind].flatten(),
                        None,
                        m,
                        h,
                    )
            else:
                raise NameError("ab_initio_property_calculator must be provided")
        return out

    @vectorize_ingredient
    def derivative_coupling_dzc(self, parameters, **kwargs):
        z = kwargs["z"]
        num_classical_coordinates = self.constants.num_classical_coordinates
        num_quantum_states = self.constants.num_quantum_states
        m = self.constants.classical_coordinate_mass
        h = self.constants.classical_coordinate_weight
        out = np.zeros(
            (num_classical_coordinates, num_quantum_states, num_quantum_states),
            dtype=complex,
        )
        traj_ind = kwargs["traj_ind"]
        has_ab_initio_property = "ab_initio_property" in parameters.keys()
        if has_ab_initio_property:
            properties = parameters["ab_initio_property"][traj_ind]
            if "derivative_coupling" in properties.keys():
                derivative_coupling_dq = properties["derivative_coupling"]
                if not (derivative_coupling_dq is None):
                    for key, val in derivative_coupling_dq.items():
                        out[:, key[0], key[1]] = (
                            dqdp_to_dzc(val.flatten(), None, m, h) / ANGSTROM_TO_BOHR
                        )  # Convert from 1/Angstrom to 1/Bohr.
                        out[:, key[1], key[0]] = -np.conj(out[:, key[0], key[1]])
                needs_derivative_coupling = False
            else:
                needs_derivative_coupling = True
        if not (has_ab_initio_property) or needs_derivative_coupling:
            property_dict = {
                "derivative_coupling": {
                    "z": z[np.newaxis],
                    "state_inds_derivative_coupling": None,
                },
            }
            ab_initio_property_calculator, has_ab_intio_property_calculator = self.get(
                "ab_initio_property_calculator"
            )
            if has_ab_intio_property_calculator:
                properties = ab_initio_property_calculator(
                    self,
                    parameters,
                    batch_size=1,
                    property_dict=property_dict,
                )[0]
                derivative_coupling_dq = properties["derivative_coupling"]
                for key, val in derivative_coupling_dq.items():
                    out[:, key[0], key[1]] = (
                        dqdp_to_dzc(val.flatten(), None, m, h) / ANGSTROM_TO_BOHR
                    )  # Convert from 1/Angstrom to 1/Bohr.
                    out[:, key[1], key[0]] = -np.conj(out[:, key[0], key[1]])
            else:
                raise NameError("ab_initio_property_calculator must be provided")
        return out

    ingredients = [
        (
            "ab_initio_property_calculator",
            ingredients.ab_initio_property_calculator_qchem,
        ),
        ("h_q", h_q),
        ("h_qc", h_qc),
        ("h_c", ingredients.h_c_free),
        ("dh_c_dzc", ingredients.dh_c_dzc_free),
        ("dh_qc_dzc", dh_qc_dzc),
        ("hop", ingredients.hop_free),
        ("init_classical", init_classical),
        ("derivative_coupling_dzc", derivative_coupling_dzc),
        ("_init_model", _init_model),
    ]



"""
This module contains the Holstein lattice Model class.
"""

import numpy as np
from qclab.model import Model
from qclab import ingredients


class HolsteinLattice(Model):
    """
    A Model class representing a nearest-neighbor tight-binding model with Holstein-type
    electron-phonon coupling with a single optical mode.

    Reference publication:
    Krotz et al. J. Chem. Phys. 2021, 154 (22), 224101. https://doi.org/10.1063/5.0053177.
    """

    def __init__(self, constants=None):
        if constants is None:
            constants = {}
        self.default_constants = {
            "kBT": 1.0,
            "g": 0.5,
            "w": 0.5,
            "N": 10,
            "J": 1.0,
            "phonon_mass": 1.0,
            "periodic": True,
        }
        super().__init__(self.default_constants, constants)
        self.update_dh_qc_dzc = False
        self.update_h_q = False

    def _init_model(self, parameters, **kwargs):
        N = self.constants.get("N")
        w = self.constants.get("w")
        phonon_mass = self.constants.get("phonon_mass")
        self.constants.num_quantum_states = N
        self.constants.num_classical_coordinates = N
        self.constants.classical_coordinate_weight = w * np.ones(N)
        self.constants.classical_coordinate_mass = phonon_mass * np.ones(N)
        return

    def _init_h_q(self, parameters, **kwargs):
        J = self.constants.get("J")
        periodic = self.constants.get("periodic")
        self.constants.nearest_neighbor_hopping_energy = J
        self.constants.nearest_neighbor_periodic = periodic
        return

    def _init_h_qc(self, parameters, **kwargs):
        N = self.constants.get("N")
        w = self.constants.get("w")
        g = self.constants.get("g")
        h = self.constants.classical_coordinate_weight
        self.constants.diagonal_linear_coupling = np.diag(
            g * w * np.sqrt(w / h) * np.ones(N)
        )
        return

    def _init_h_c(self, parameters, **kwargs):
        N = self.constants.get("N")
        w = self.constants.get("w")
        self.constants.harmonic_frequency = w * np.ones(N)
        return

    ingredients = [
        ("h_q", ingredients.h_q_nearest_neighbor),
        ("h_qc", ingredients.h_qc_diagonal_linear),
        ("h_c", ingredients.h_c_harmonic),
        ("dh_qc_dzc", ingredients.dh_qc_dzc_diagonal_linear),
        ("dh_c_dzc", ingredients.dh_c_dzc_harmonic),
        ("init_classical", ingredients.init_classical_boltzmann_harmonic),
        ("hop", ingredients.hop_harmonic),
        ("_init_h_q", _init_h_q),
        ("_init_h_qc", _init_h_qc),
        ("_init_h_c", _init_h_c),
        ("_init_model", _init_model),
    ]


class HolsteinLatticeReciprocalSpace(Model):
    """
    A model representing a nearest-neighbor tight-binding model with Holstein-type
    electron-phonon coupling with a single optical mode implemented in reciprocal space.

    Reference publication:
    Krotz et al. J. Chem. Phys. 154, 224101 (2021); https://doi.org/10.1063/5.0053177
    """

    def __init__(self, constants=None):
        if constants is None:
            constants = {}
        self.default_constants = {
            "kBT": 1.0,
            "g": 0.5,
            "w": 0.5,
            "N": 10,
            "J": 1.0,
            "phonon_mass": 1.0,
            "periodic": True,
        }
        super().__init__(self.default_constants, constants)
        self.update_dh_qc_dzc = False
        self.update_h_q = False

    def init_model(self, parameters, **kwargs):
        N = self.constants.get("N")
        assert N % 2 == 0, "N must be even."
        self.constants.num_quantum_states = N
        self.constants.num_classical_coordinates = N
        self.constants.classical_coordinate_weight = self.constants.get("w") * np.ones(
            N
        )
        self.constants.classical_coordinate_mass = self.constants.get(
            "phonon_mass"
        ) * np.ones(N)
        self.constants.harmonic_frequency = self.constants.get("w") * np.ones(N)
        self.constants.k_inds = np.arange(-N / 2, N / 2).astype(int)
        self.constants.k_diff_inds = np.array(
            [np.roll(self.constants.k_inds, int(N / 2 + i)) for i in range(N)]
        ).astype(int)
        return

    def h_q(self, parameters, **kwargs):
        batch_size = kwargs["batch_size"]
        J = self.constants.get("J")
        k = 2.0 * np.pi * self.constants.k_inds / self.constants.num_quantum_states
        out = np.diag(-2.0 * J * np.cos(k))
        h_q = np.broadcast_to(
            out[np.newaxis, :, :],
            (
                batch_size,
                self.constants.num_quantum_states,
                self.constants.num_quantum_states,
            ),
        )
        return h_q

    def h_qc(self, parameters, **kwargs):
        z = kwargs["z"]
        g = self.constants.get("g")
        w = self.constants.get("w")
        h = self.constants.classical_coordinate_weight
        z_kap_mat = z[:, self.constants.k_diff_inds]
        zc_mkap_mat = np.conj(z[:, self.constants.k_diff_inds.transpose()])
        h_qc = (g * w * np.sqrt(w / h) / np.sqrt(self.constants.num_quantum_states)) * (
            z_kap_mat + zc_mkap_mat
        )
        return h_qc

    def dh_qc_dzc(self, parameters, **kwargs):
        z = kwargs["z"]
        batch_size = len(z)
        g = self.constants.get("g")
        w = self.constants.get("w")
        h = self.constants.classical_coordinate_weight
        out = np.zeros(
            (
                batch_size,
                self.constants.num_classical_coordinates,
                self.constants.num_quantum_states,
                self.constants.num_quantum_states,
            ),
            dtype=complex,
        )
        for k_ind in self.constants.k_inds:
            pos = np.where(self.constants.k_diff_inds.transpose() == k_ind)
            out[:, k_ind, pos[0], pos[1]] = (
                g * w * np.sqrt(w / h) / np.sqrt(self.constants.num_quantum_states)
            )
        shape = (
            batch_size,
            self.constants.num_classical_coordinates,
            self.constants.num_quantum_states,
            self.constants.num_quantum_states,
        )
        inds = np.where(out != 0)
        mels = out[inds]
        return inds, mels, shape

    ingredients = [
        ("h_q", h_q),
        ("h_qc", h_qc),
        ("h_c", ingredients.h_c_harmonic),
        ("dh_qc_dzc", dh_qc_dzc),
        ("dh_c_dzc", ingredients.dh_c_dzc_harmonic),
        ("init_classical", ingredients.init_classical_boltzmann_harmonic),
        ("hop", ingredients.hop_harmonic),
        ("_init_model", init_model),
    ]



"""
This module contains the spin-boson Model class.
"""

import numpy as np
from qclab.model import Model
from qclab import ingredients


class SpinBoson(Model):
    """
    Spin-Boson Model class.

    Reference publication:
    Tempelaar & Reichman. J. Chem. Phys. 2018, 148 (10), 102309. https://doi.org/10.1063/1.5000843.
    """

    def __init__(self, constants=None):
        if constants is None:
            constants = {}
        self.default_constants = {
            "kBT": 1.0,
            "V": 0.5,
            "E": 0.5,
            "A": 100,
            "W": 0.1,
            "l_reorg": 0.005,
            "boson_mass": 1.0,
        }
        super().__init__(self.default_constants, constants)
        self.update_dh_qc_dzc = False
        self.update_h_q = False

    def _init_h_q(self, parameters, **kwargs):
        self.constants.two_level_00 = self.constants.get("E")
        self.constants.two_level_11 = -self.constants.get("E")
        self.constants.two_level_01_re = self.constants.get("V")
        self.constants.two_level_01_im = 0
        return

    def _init_h_qc(self, parameters, **kwargs):
        A = self.constants.get("A")
        l_reorg = self.constants.get("l_reorg")
        boson_mass = self.constants.get("boson_mass")
        h = self.constants.classical_coordinate_weight
        w = self.constants.harmonic_frequency
        self.constants.diagonal_linear_coupling = np.zeros((2, A))
        self.constants.diagonal_linear_coupling[0] = (
            w * np.sqrt(2.0 * l_reorg / A) * (1.0 / np.sqrt(2.0 * boson_mass * h))
        )
        self.constants.diagonal_linear_coupling[1] = (
            -w * np.sqrt(2.0 * l_reorg / A) * (1.0 / np.sqrt(2.0 * boson_mass * h))
        )
        return

    def _init_h_c(self, parameters, **kwargs):
        A = self.constants.get("A")
        W = self.constants.get("W")
        self.constants.harmonic_frequency = W * np.tan(
            np.arange(0.5, A + 0.5, 1.0) * np.pi * 0.5 / A
        )
        return

    def _init_model(self, parameters, **kwargs):
        A = self.constants.get("A")
        boson_mass = self.constants.get("boson_mass")
        self.constants.num_classical_coordinates = A
        self.constants.num_quantum_states = 2
        self.constants.classical_coordinate_weight = self.constants.harmonic_frequency
        self.constants.classical_coordinate_mass = boson_mass * np.ones(A)
        return

    ingredients = [
        ("h_q", ingredients.h_q_two_level),
        ("h_qc", ingredients.h_qc_diagonal_linear),
        ("h_c", ingredients.h_c_harmonic),
        ("dh_qc_dzc", ingredients.dh_qc_dzc_diagonal_linear),
        ("dh_c_dzc", ingredients.dh_c_dzc_harmonic),
        ("init_classical", ingredients.init_classical_boltzmann_harmonic),
        ("hop", ingredients.hop_harmonic),
        ("_init_h_q", _init_h_q),
        ("_init_h_qc", _init_h_qc),
        ("_init_model", _init_model),
        ("_init_h_c", _init_h_c),
    ]



"""
This module contains the Data class.
"""

import logging
import numpy as np
from qclab.utils import DISABLE_H5PY

if not DISABLE_H5PY:
    import h5py

logger = logging.getLogger(__name__)


class Data:
    """
    Data class for handling the collection of data during a simulation.
    """

    def __init__(self, seeds=None):
        if seeds is None:
            seeds = np.array([], dtype=int)
        self.data_dict = {"seed": seeds, "norm_factor": 0}
        # Store log messages captured during a simulation run. This attribute is
        # populated by the drivers when they return the Data object.
        self.log = ""

    def add_output_to_data_dict(self, sim, state, t_ind):
        """
        Add data to the output dictionary ``self.data_dict``.

        .. rubric:: Args
        sim: Simulation
            The Simulation object containing settings and parameters.
        state: dict
            The dictionary containing the current simulation state.
        t_ind: int
            The current time index in the simulation.
        """
        # Check if the norm_factor is zero. If it is, save it from the State object.
        if self.data_dict["norm_factor"] == 0:
            if "norm_factor" not in state:
                logger.critical(
                    "The State object does not have a norm_factor attribute. "
                    "This is required to normalize the data."
                )
                raise ValueError(
                    "The State object does not have a norm_factor attribute."
                )
            logger.info(
                "Setting norm_factor to %s from State object.", state["norm_factor"]
            )
            self.data_dict["norm_factor"] = state["norm_factor"]
        for key, val in state["output_dict"].items():
            if not key in self.data_dict:
                # If the key is not in the data_dict, initialize it with zeros.
                self.data_dict[key] = np.zeros(
                    (len(sim.settings.t_collect), *np.shape(val)[1:]), dtype=val.dtype
                )
                logger.info(
                    "Initializing data_dict[%s] with shape %s.",
                    key,
                    self.data_dict[key].shape,
                )
            # Store the data in the data_dict at the correct time index.
            self.data_dict[key][t_ind // sim.settings.dt_collect_n] = (
                np.sum(val, axis=0) / self.data_dict["norm_factor"]
            )

    def add_data(self, new_data):
        """
        Add data from ``new_data`` to the output dictionary ``self.data_dict``.

        .. rubric:: Args
        new_data: Data
            A Data instance containing the new data to merge.
        """
        new_norm_factor = (
            new_data.data_dict["norm_factor"] + self.data_dict["norm_factor"]
        )
        for key, val in new_data.data_dict.items():
            if key == "seed":
                self.data_dict[key] = np.concatenate(
                    (self.data_dict[key], val.flatten()), axis=0
                )
            elif key != "norm_factor":
                if key in self.data_dict:
                    self.data_dict[key] = (
                        self.data_dict[key] * self.data_dict["norm_factor"]
                        + val * new_data.data_dict["norm_factor"]
                    ) / new_norm_factor
                else:
                    self.data_dict[key] = val
        self.data_dict["norm_factor"] = new_norm_factor
        # Append any log messages stored in new_data to this instance's log.
        if getattr(new_data, "log", ""):
            self.log += new_data.log

    def save(self, filename, disable_h5py=DISABLE_H5PY):
        """
        Save the data to disk with file name ``filename``.

        If h5py is available the data is stored as an HDF5 archive; otherwise
        each variable is saved using numpy.savez.

        .. rubric:: Args
        filename : str
            The file name to save the data to.
        disable_h5py : bool, default: qclab.utils.DISABLE_H5PY
            If True, h5py is not used even if available.
        """
        if disable_h5py:
            np.savez(filename, log=self.log, **self.data_dict)
        else:
            with h5py.File(filename, "w") as h5file:
                self._recursive_save(h5file, "/", self.data_dict)
                h5file.attrs["log"] = self.log

    def load(self, filename, disable_h5py=DISABLE_H5PY):
        """
        Load a Data object from ``filename``.

        .. rubric:: Args
        filename : str
            The file name to load the data from.
        disable_h5py : bool, default: qclab.utils.DISABLE_H5PY
            If True, h5py is not used even if available.

        .. rubric:: Returns
        Data : Data
            The loaded Data object.
        """
        new_data = Data()
        if disable_h5py:
            loaded = np.load(filename, allow_pickle=True)
            new_data.data_dict = {
                key: loaded[key] for key in loaded.files if key != "log"
            }
            new_data.log = str(loaded.get("log", ""))
            self.add_data(new_data)
            return self
        with h5py.File(filename, "r") as h5file:
            new_data._recursive_load(h5file, "/", new_data.data_dict)
            new_data.log = h5file.attrs["log"]
            self.add_data(new_data)
        return self

    def _recursive_save(self, h5file, path, dict):
        """
        Recursively save dictionary contents to an HDF5 group.

        .. rubric:: Args
        h5file : h5py.File
            The HDF5 file object to save data into.
        path : str
            The current path in the HDF5 file.
        dict : dict
            The dictionary to save.
        """
        for key, item in dict.items():
            if isinstance(item, type(dict)):
                self._recursive_save(h5file, path + key + "/", item)
            else:
                try:
                    h5file.create_dataset(path + key, data=np.asarray(item))
                except TypeError as exc:
                    raise ValueError(
                        f"Cannot save {key} with unsupported type {type(item)}."
                    ) from exc

    def _recursive_load(self, h5file, path, dict):
        """
        Recursively load dictionary contents from an HDF5 group.

        .. rubric:: Args
        h5file : h5py.File
            The HDF5 file object to load data from.
        path : str
            The current path in the HDF5 file.
        dict : dict
            The dictionary to populate with loaded data.
        """
        for key, item in h5file[path].items():
            if isinstance(item, h5py._hl.dataset.Dataset):
                dict[key] = item[()]
            elif isinstance(item, h5py._hl.group.Group):
                dict[key] = {}
                self._recursive_load(h5file, path + key + "/", dict[key])



"""
This module contains numerical constants.
"""

# Numerical threshold for near zero values.
SMALL = 1e-10

# Numerical threshold for nonadiabatic coupling gauge fixing.
# The misalignment is allowed to be 1e-3 of the magnitude of the coupling.
GAUGE_FIX_THRESHOLD = 1e-3

# Finite difference step size.
FINITE_DIFFERENCE_DELTA = 1e-6

# Conversion factor from electronvolts to wavenumbers.
EV_TO_INVCM = 8065.610420

# Conversion factor from Hartrees to electronvolts.
HA_TO_EV = 27.21138625

# Conversion factor from eV to Hartrees.
EV_TO_HA = 1 / HA_TO_EV

# Conversion factor from wavenumbers to Hartrees.
INVCM_TO_HA = (1 / EV_TO_INVCM) * EV_TO_HA

# Conversion factor from Angstroms to Bohr.
ANGSTROM_TO_BOHR = 1.8897259886

# Conversion factor from atomic mass units to electron mass.
AMU_TO_EMASS = 1822.89

# Conversion factor from atomic units of time to femtoseconds.
AU_TIME_TO_FS = 0.02419


"""
This module contains numerical constants.

Values for physical constants are taken from 2022 CODATA recommended values.
Reference publication:
Mohr et al. Rev. Mod. Phys. 2025, 97 (2), 025002. https://doi.org/10.1103/RevModPhys.97.025002.
"""

import numpy as np

# Numerical threshold for near zero values.
SMALL = 1e-10

# Numerical threshold for nonadiabatic coupling gauge fixing.
# The misalignment is allowed to be 1e-3 of the magnitude of the coupling.
GAUGE_FIX_THRESHOLD = 1e-3

# Finite difference step size.
FINITE_DIFFERENCE_DELTA = 1e-6

# Speed of light [m/s].
C_M_PER_S = 299792458
# Planck constant [J * s].
H_J_S = 6.62607015e-34
# Boltzmann constant [J/K].
K_B_J_PER_K = 1.380649e-23

# Reduced Planck constant [J·s].
HBAR_J_S = H_J_S / (2 * np.pi)

# Reference temperature [K].
T_REF_K = 300

# Thermal energy at the reference temperature [J].
KBT_REF_J = K_B_J_PER_K * T_REF_K

# h * c [J*m].
HC_J_M = H_J_S * C_M_PER_S

# Conversion between inverse centimeters to reference energy.
# kBT / invcm = (100 [cm / m]) * hc[J*m] / kBT [J]
# A[INVCM] * INVCM_TO_KBT_REF = A[KBT_REF]
INVCM_TO_KBT_REF = 100 * HC_J_M / KBT_REF_J

# Alias to old name for backwards-compatibility.
INVCM_TO_300K = INVCM_TO_KBT_REF



"""
This module contains the Model class for the Fenna-Matthews-Olson (FMO) complex.
"""

import numpy as np
from qclab.model import Model
from qclab import ingredients
from qclab.numerical_constants import INVCM_TO_300K


class FMOComplex(Model):
    """
    A Model class representing the Fenna-Matthews-Olson (FMO) complex.

    All quantities in this model are taken to be in units of kBT at 300 K.

    At 300 K, kBT = 0.025852 eV = 208.521 cm^-1. Any quantity in wavenumbers
    is made unitless by dividing by kBT.

    One unit of time is given by \\hbar/kBT = 25.4595 fs.

    Reference publication:
    Mulvihill et al. J. Chem. Phys. 2021, 154 (20), 204109. https://doi.org/10.1063/5.0051101.
    """

    def __init__(self, constants=None):
        if constants is None:
            constants = {}
        self.default_constants = {
            "kBT": 1.0,
            "mass": 1.0,
            "l_reorg": 35.0 * INVCM_TO_300K,  # Reorganization energy
            "w_c": 106.14 * INVCM_TO_300K,  # Characteristic frequency
            "N": 200,
        }
        super().__init__(self.default_constants, constants)
        self.update_dh_qc_dzc = False
        self.update_h_q = False

    def _init_model(self, parameters, **kwargs):
        """
        Initialize the model-specific constants.
        """
        self.constants.num_quantum_states = 7
        N = self.constants.get("N")
        w_c = self.constants.get("w_c")
        mass = self.constants.get("mass")
        self.constants.w = (
            w_c
            * np.tan(np.arange(0.5, N + 0.5, 1.0) * np.pi * 0.5 / (N))[np.newaxis, :]
            * np.ones((self.constants.num_quantum_states, N))
        ).flatten()
        self.constants.num_classical_coordinates = self.constants.num_quantum_states * N

        self.constants.classical_coordinate_weight = self.constants.w
        self.constants.classical_coordinate_mass = mass * np.ones(
            self.constants.num_classical_coordinates
        )
        return

    def _init_h_c(self, parameters, **kwargs):
        self.constants.harmonic_frequency = self.constants.w
        return

    def _init_h_qc(self, parameters, **kwargs):
        N = self.constants.get("N")
        l_reorg = self.constants.get("l_reorg")
        m = self.constants.classical_coordinate_mass
        h = self.constants.classical_coordinate_weight
        w = self.constants.w
        self.constants.diagonal_linear_coupling = np.zeros(
            (
                self.constants.num_quantum_states,
                self.constants.num_classical_coordinates,
            )
        )
        for n in range(self.constants.num_quantum_states):
            self.constants.diagonal_linear_coupling[n, n * N : (n + 1) * N] = (
                w * np.sqrt(2.0 * l_reorg / N) * (1.0 / np.sqrt(2.0 * m * h))
            )[n * N : (n + 1) * N]
        return

    def h_q(self, parameters, **kwargs):
        batch_size = kwargs["batch_size"]
        matrix_elements = (
            np.array(
                [
                    [12410, -87.7, 5.5, -5.9, 6.7, -13.7, -9.9],
                    [-87.7, 12530, 30.8, 8.2, 0.7, 11.8, 4.3],
                    [5.5, 30.8, 12210.0, -53.5, -2.2, -9.6, 6.0],
                    [-5.9, 8.2, -53.5, 12320, -70.7, -17.0, -63.3],
                    [6.7, 0.7, -2.2, -70.7, 12480, 81.1, -1.3],
                    [-13.7, 11.8, -9.6, -17.0, 81.1, 12630, 39.7],
                    [-9.9, 4.3, 6.0, -63.3, -1.3, 39.7, 12440],
                ],
                dtype=complex,
            )
            * INVCM_TO_300K
        )
        # To reduce numerical errors we offset the diagonal elements by
        # their minimum value.
        matrix_elements = matrix_elements - np.min(
            np.diag(matrix_elements)
        ) * np.identity(7)
        # Finally we broadcast the array to the desired shape.
        return np.broadcast_to(matrix_elements, (batch_size, 7, 7))

    ingredients = [
        ("h_q", h_q),
        ("h_qc", ingredients.h_qc_diagonal_linear),
        ("h_c", ingredients.h_c_harmonic),
        ("dh_qc_dzc", ingredients.dh_qc_dzc_diagonal_linear),
        ("dh_c_dzc", ingredients.dh_c_dzc_harmonic),
        ("init_classical", ingredients.init_classical_boltzmann_harmonic),
        ("hop", ingredients.hop_harmonic),
        ("_init_h_qc", _init_h_qc),
        ("_init_h_c", _init_h_c),
        ("_init_model", _init_model),
    ]



"""
This module runs the models and algorithms in QC Lab for a given simulation setting and
compares them against a reference dataset.
"""

import pytest
import os
import time
import numpy as np
from qclab import Simulation, Data
from qclab.models import (
    SpinBoson,
    HolsteinLattice,
    HolsteinLatticeReciprocalSpace,
    FMOComplex,
    TullyProblemOne,
    TullyProblemTwo,
    TullyProblemThree,
)
from qclab.algorithms import MeanField, FewestSwitchesSurfaceHopping
from qclab.dynamics import serial_driver, parallel_driver_multiprocessing

try:
    from tests.reference_settings import model_sim_settings, model_settings
except ImportError:
    from reference_settings import model_sim_settings, model_settings


def test_output_serial():
    """
    This test verifies that the serial driver produces the expected results.
    """
    reference_folder = os.path.join(os.path.dirname(__file__), "reference/")
    for model_class in [
        SpinBoson,
        HolsteinLattice,
        HolsteinLatticeReciprocalSpace,
        FMOComplex,
        TullyProblemOne,
        TullyProblemTwo,
        TullyProblemThree,
    ]:
        for algorithm_class in [MeanField, FewestSwitchesSurfaceHopping]:
            print(f"Testing {model_class.__name__} with {algorithm_class.__name__}")

            sim = Simulation(model_sim_settings[model_class.__name__])
            model_name = model_class.__name__
            algorithm_name = algorithm_class.__name__
            sim.model = model_class(model_settings[model_class.__name__])
            sim.model.initialize_constants()
            sim.algorithm = algorithm_class()
            sim.initial_state["wf_db"] = np.zeros(
                sim.model.constants.num_quantum_states, dtype=complex
            )
            sim.initial_state["wf_db"][0] = 1j
            st = time.time()
            data = serial_driver(sim)
            et = time.time()
            print(f"Finished in {et - st:.2f} seconds.")
            data_correct = Data().load(
                os.path.join(reference_folder, f"{model_name}_{algorithm_name}.h5")
            )
            for key, val in data.data_dict.items():
                np.testing.assert_allclose(
                    val, data_correct.data_dict[key], rtol=1e-5, atol=1e-8
                )
    return


def test_output_multiprocessing():
    """
    This test verifies that the multiprocessing driver produces the same results
    as the serial driver.
    """
    reference_folder = os.path.join(os.path.dirname(__file__), "reference/")
    for model_class in [
        SpinBoson,
        HolsteinLattice,
        HolsteinLatticeReciprocalSpace,
        FMOComplex,
        TullyProblemOne,
        TullyProblemTwo,
        TullyProblemThree,
    ]:
        for algorithm_class in [MeanField, FewestSwitchesSurfaceHopping]:
            print(f"Testing {model_class.__name__} with {algorithm_class.__name__}")

            sim = Simulation(model_sim_settings[model_class.__name__])
            sim.settings.batch_size = sim.settings.num_trajs // 2
            model_name = model_class.__name__
            algorithm_name = algorithm_class.__name__
            sim.model = model_class(model_settings[model_class.__name__])
            sim.model.initialize_constants()
            sim.algorithm = algorithm_class()
            sim.initial_state["wf_db"] = np.zeros(
                sim.model.constants.num_quantum_states, dtype=complex
            )
            sim.initial_state["wf_db"][0] = 1j
            st = time.time()
            data = parallel_driver_multiprocessing(sim)
            et = time.time()
            print(f"Finished in {et - st:.2f} seconds.")
            data_correct = Data().load(
                os.path.join(reference_folder, f"{model_name}_{algorithm_name}.h5")
            )
            for key, val in data.data_dict.items():
                np.testing.assert_allclose(
                    val, data_correct.data_dict[key], rtol=1e-5, atol=1e-8
                )
    return


def test_output_different_h():
    """
    This test ensures that changing the value of the ``classical_coordinate_weight`` does
    not change the result of the simulation.
    """
    reference_folder = os.path.join(os.path.dirname(__file__), "reference/")
    for model_class in [
        SpinBoson,
        HolsteinLattice,
        HolsteinLatticeReciprocalSpace,
        FMOComplex,
        TullyProblemOne,
        TullyProblemTwo,
        TullyProblemThree,
    ]:
        for algorithm_class in [MeanField, FewestSwitchesSurfaceHopping]:
            print(f"Testing {model_class.__name__} with {algorithm_class.__name__}")

            sim = Simulation(model_sim_settings[model_class.__name__])
            model_name = model_class.__name__
            algorithm_name = algorithm_class.__name__
            sim.model = model_class(model_settings[model_class.__name__])
            sim.model.initialize_constants()
            np.random.seed(10)
            # Adjust the weights to be a random positive number between 1 and 2.
            sim.model.constants.classical_coordinate_weight = (
                np.random.rand(sim.model.constants.num_classical_coordinates) + 1.0
            )
            sim.algorithm = algorithm_class()
            sim.initial_state["wf_db"] = np.zeros(
                sim.model.constants.num_quantum_states, dtype=complex
            )
            sim.initial_state["wf_db"][0] = 1j
            st = time.time()
            data = serial_driver(sim)
            et = time.time()
            print(f"Finished in {et - st:.2f} seconds.")
            data_correct = Data().load(
                os.path.join(reference_folder, f"{model_name}_{algorithm_name}.h5")
            )
            for key, val in data.data_dict.items():
                np.testing.assert_allclose(
                    val, data_correct.data_dict[key], rtol=1e-5, atol=1e-8
                )
    return


def test_output_fssh_gauge_fixing():
    """
    This test ensures that the initial choice of phase in FSSH does not affect
    the final result of a calculation.

    The test proceeds by adding a random phase to the eigenvectors computed at the
    outset of the simulation. The relative phases of these vectors are fixed so that
    the derivative couplings are real-valued ("phase_der_couple"). At subsequent
    timesteps, the relative phase of the new eigenvectors is determined from maximizing the
    real part of their overlap with the previous set of eigenvectors ("phase_overlap").

    Note that this test will only work if the ``add_random_phase`` task is between the
    ``diagonalize_matrix`` and ``update_eigvecs_gauge`` tasks in the initialization recipe.

    """
    reference_folder = os.path.join(os.path.dirname(__file__), "reference/")
    my_FSSH = FewestSwitchesSurfaceHopping({"gauge_fixing": "phase_overlap"})

    def add_random_phase(sim, state, parameters):
        np.random.seed(10)
        random_phases = np.exp(
            1j
            * 2.0
            * np.pi
            * np.random.rand(
                sim.settings.batch_size, sim.model.constants.num_quantum_states
            )
        )
        state["eigvecs"] = random_phases[:, np.newaxis, :] * state["eigvecs"]
        return state, parameters

    my_FSSH.initialization_recipe.insert(6, add_random_phase)
    for model_class in [
        SpinBoson,
        HolsteinLattice,
        HolsteinLatticeReciprocalSpace,
        FMOComplex,
        TullyProblemOne,
        TullyProblemTwo,
        TullyProblemThree,
    ]:
        for algorithm_class in [my_FSSH]:
            print(f"Testing {model_class.__name__} with FewestSwitchesSurfaceHopping")
            sim = Simulation(model_sim_settings[model_class.__name__])
            model_name = model_class.__name__
            algorithm_name = "FewestSwitchesSurfaceHopping"
            sim.model = model_class(model_settings[model_class.__name__])
            sim.model.initialize_constants()
            sim.algorithm = algorithm_class
            sim.initial_state["wf_db"] = np.zeros(
                sim.model.constants.num_quantum_states, dtype=complex
            )
            sim.initial_state["wf_db"][0] = 1j
            st = time.time()
            data = serial_driver(sim)
            et = time.time()
            print(f"Finished in {et - st:.2f} seconds.")
            data_correct = Data().load(
                os.path.join(reference_folder, f"{model_name}_{algorithm_name}.h5")
            )
            for key, val in data.data_dict.items():
                np.testing.assert_allclose(
                    val, data_correct.data_dict[key], rtol=1e-5, atol=1e-8
                )
    return


def test_output_fssh_deterministic():
    """
    This test verifies the deterministic version of FSSH.
    """
    reference_folder = os.path.join(os.path.dirname(__file__), "reference/")
    for model_class in [
        SpinBoson,
        HolsteinLattice,
        HolsteinLatticeReciprocalSpace,
        FMOComplex,
        TullyProblemOne,
        TullyProblemTwo,
        TullyProblemThree,
    ]:
        algorithm_class = FewestSwitchesSurfaceHopping
        print(
            f"Testing {model_class.__name__} with deterministic {algorithm_class.__name__}"
        )

        sim = Simulation(model_sim_settings[model_class.__name__])
        model_name = model_class.__name__
        algorithm_name = algorithm_class.__name__
        sim.model = model_class(model_settings[model_class.__name__])
        sim.model.initialize_constants()
        sim.settings.batch_size *= sim.model.constants.num_quantum_states
        sim.settings.num_trajs *= sim.model.constants.num_quantum_states
        sim.algorithm = algorithm_class({"fssh_deterministic": True})
        sim.initial_state["wf_db"] = np.zeros(
            sim.model.constants.num_quantum_states, dtype=complex
        )
        sim.initial_state["wf_db"][0] = 1j
        st = time.time()
        data = serial_driver(sim)
        et = time.time()
        print(f"Finished in {et - st:.2f} seconds.")
        data_correct = Data().load(
            os.path.join(
                reference_folder, f"{model_name}_{algorithm_name}_deterministic.h5"
            )
        )
        for key, val in data.data_dict.items():
            np.testing.assert_allclose(
                val, data_correct.data_dict[key], rtol=1e-5, atol=1e-8
            )
    return


def test_dh_qc_dzc_finite_differences():
    """
    Tests the output when using finite differences to compute dh_qc_dzc.

    We change the atol to 1e-5 so that it matches better in TullyProblemThree.
    """
    reference_folder = os.path.join(os.path.dirname(__file__), "reference/")
    for model_class in [
        SpinBoson,
        HolsteinLattice,
        HolsteinLatticeReciprocalSpace,
        FMOComplex,
        TullyProblemOne,
        TullyProblemTwo,
        TullyProblemThree,
    ]:
        for algorithm_class in [MeanField, FewestSwitchesSurfaceHopping]:
            print(f"Testing {model_class.__name__} with {algorithm_class.__name__}")

            sim = Simulation(model_sim_settings[model_class.__name__])
            model_name = model_class.__name__
            algorithm_name = algorithm_class.__name__
            sim.model = model_class(model_settings[model_class.__name__])
            sim.model.ingredients.append(("dh_qc_dzc", None))
            sim.model.initialize_constants()
            sim.algorithm = algorithm_class()
            sim.initial_state["wf_db"] = np.zeros(
                sim.model.constants.num_quantum_states, dtype=complex
            )
            sim.initial_state["wf_db"][0] = 1j
            st = time.time()
            data = serial_driver(sim)
            et = time.time()
            print(f"Finished in {et - st:.2f} seconds.")
            data_correct = Data().load(
                os.path.join(reference_folder, f"{model_name}_{algorithm_name}.h5")
            )
            for key, val in data.data_dict.items():
                np.testing.assert_allclose(
                    val, data_correct.data_dict[key], rtol=1e-5, atol=1e-5
                )
    return


def test_dh_c_dzc_finite_differences():
    """
    Tests the output when using finite differences to compute dh_c_dzc.
    We change the atol to 1e-4 and reduce the finite difference delta to 1e-9.
    This is just to obtain agreement with reference data,
    the defaults may be appropriate in other cases.

    We exclude SpinBoson and FMOComplex here because they take too long.
    """
    reference_folder = os.path.join(os.path.dirname(__file__), "reference/")
    for model_class in [
        # SpinBoson,
        HolsteinLattice,
        HolsteinLatticeReciprocalSpace,
        # FMOComplex,
        TullyProblemOne,
        TullyProblemTwo,
        TullyProblemThree,
    ]:
        for algorithm_class in [MeanField, FewestSwitchesSurfaceHopping]:
            print(f"Testing {model_class.__name__} with {algorithm_class.__name__}")

            sim = Simulation(model_sim_settings[model_class.__name__])
            model_name = model_class.__name__
            algorithm_name = algorithm_class.__name__
            sim.model = model_class(model_settings[model_class.__name__])
            sim.model.ingredients.append(("dh_c_dzc", None))
            sim.model.initialize_constants()
            sim.model.constants.dh_c_dzc_finite_difference_delta = 1e-9
            sim.algorithm = algorithm_class()
            sim.initial_state["wf_db"] = np.zeros(
                sim.model.constants.num_quantum_states, dtype=complex
            )
            sim.initial_state["wf_db"][0] = 1j
            st = time.time()
            data = serial_driver(sim)
            et = time.time()
            print(f"Finished in {et - st:.2f} seconds.")
            data_correct = Data().load(
                os.path.join(reference_folder, f"{model_name}_{algorithm_name}.h5")
            )
            for key, val in data.data_dict.items():
                np.testing.assert_allclose(
                    val, data_correct.data_dict[key], rtol=1e-5, atol=1e-4
                )
    return


if __name__ == "__main__":
    st = time.time()
    test_output_serial()
    et1 = time.time()
    print(f"Serial tests completed in {et1 - st:.2f} seconds.")
    test_output_multiprocessing()
    et2 = time.time()
    print(f"Multiprocessing tests completed in {et2 - et1:.2f} seconds.")
    test_output_different_h()
    et3 = time.time()
    print(f"Different h tests completed in {et3 - et2:.2f} seconds.")
    test_output_fssh_gauge_fixing()
    et4 = time.time()
    print(f"FSSH gauge fixing tests completed in {et4 - et3:.2f} seconds.")
    test_output_fssh_deterministic()
    et5 = time.time()
    print(f"FSSH deterministic tests completed in {et5 - et4:.2f} seconds.")
    test_dh_qc_dzc_finite_differences()
    et6 = time.time()
    print(f"dh_qc_dzc finite difference tests completed in {et6 - et5:.2f} seconds.")
    test_dh_c_dzc_finite_differences()
    et7 = time.time()
    print(f"dh_c_dzc finite difference tests completed in {et7 - et6:.2f} seconds.")
    print(f"All tests completed in {et7 - st:.2f} seconds.")



"""
This module contains tests of the data object methods.
"""

import pytest


def test_save_load_h5py():
    import numpy as np
    from qclab import Simulation, Data
    from qclab.models import SpinBoson
    from qclab.algorithms import MeanField
    from qclab.dynamics import serial_driver
    import os

    try:
        import h5py as _
    except ImportError:
        pytest.skip("h5py not available, skipping test.", allow_module_level=True)

    sim = Simulation()
    sim.settings.progress_bar = False
    sim.settings.num_trajs = 200
    sim.settings.batch_size = 50
    sim.settings.tmax = 10
    sim.settings.dt_update = 0.01

    sim.model = SpinBoson(
        {
            "V": 0.5,
            "E": 0.5,
            "A": 100,
            "W": 0.1,
            "l_reorg": 0.005,
            "boson_mass": 1.0,
            "kBT": 1.0,
        }
    )
    sim.algorithm = MeanField()
    sim.initial_state["wf_db"] = np.zeros(
        (sim.model.constants.num_quantum_states), dtype=complex
    )
    sim.initial_state["wf_db"][0] += 1.0
    data_serial = serial_driver(sim)
    data_serial.save("test_data.h5")
    loaded_data = Data().load("test_data.h5")
    print("Comparing results...")
    for key, val in data_serial.data_dict.items():
        if isinstance(val, np.ndarray):
            assert np.allclose(val, loaded_data.data_dict[key])
    print("calculated and loaded results match!")
    os.remove("test_data.h5")
    return


def test_save_load_no_h5py():
    import numpy as np
    from qclab import Simulation, Data
    from qclab.models import SpinBoson
    from qclab.algorithms import MeanField
    from qclab.dynamics import serial_driver
    import os

    sim = Simulation()
    sim.settings.progress_bar = False
    sim.settings.num_trajs = 200
    sim.settings.batch_size = 50
    sim.settings.tmax = 10
    sim.settings.dt_update = 0.01

    sim.model = SpinBoson(
        {
            "V": 0.5,
            "E": 0.5,
            "A": 100,
            "W": 0.1,
            "l_reorg": 0.005,
            "boson_mass": 1.0,
            "kBT": 1.0,
        }
    )
    sim.algorithm = MeanField()
    sim.initial_state["wf_db"] = np.zeros(
        (sim.model.constants.num_quantum_states), dtype=complex
    )
    sim.initial_state["wf_db"][0] += 1.0
    data_serial = serial_driver(sim)
    data_serial.save("test_data.npz", disable_h5py=True)
    loaded_data = Data().load("test_data.npz", disable_h5py=True)
    print("Comparing results...")
    for key, val in data_serial.data_dict.items():
        if isinstance(val, np.ndarray):
            assert np.allclose(val, loaded_data.data_dict[key])
    print("calculated and loaded results match!")
    os.remove("test_data.npz")
    return


def test_load_sum():
    import numpy as np
    from qclab import Simulation, Data
    from qclab.models import SpinBoson
    from qclab.algorithms import MeanField
    from qclab.dynamics import serial_driver
    import os

    try:
        import h5py as _
    except ImportError:
        pytest.skip("h5py not available, skipping test.", allow_module_level=True)

    sim = Simulation()
    sim.settings.progress_bar = False
    sim.settings.num_trajs = 200
    sim.settings.batch_size = 50
    sim.settings.tmax = 10
    sim.settings.dt_update = 0.01

    sim.model = SpinBoson(
        {
            "V": 0.5,
            "E": 0.5,
            "A": 100,
            "W": 0.1,
            "l_reorg": 0.005,
            "boson_mass": 1.0,
            "kBT": 1.0,
        }
    )
    sim.algorithm = MeanField()
    sim.initial_state["wf_db"] = np.zeros(
        (sim.model.constants.num_quantum_states), dtype=complex
    )
    sim.initial_state["wf_db"][0] += 1.0
    data_serial = serial_driver(sim)
    data_serial.save("test_data.h5")
    new_data = Data().load("test_data.h5")
    new_data = new_data.load("test_data.h5")
    assert np.all(
        new_data.data_dict["seed"]
        == np.concatenate(
            (data_serial.data_dict["seed"], data_serial.data_dict["seed"])
        )
    )
    print("Comparing results...")
    for key, val in data_serial.data_dict.items():
        if isinstance(val, np.ndarray):
            if key == "seed":
                continue
            assert np.allclose(val, new_data.data_dict[key])
    print("calculated and loaded results match!")
    os.remove("test_data.h5")


if __name__ == "__main__":
    test_save_load_h5py()
    test_save_load_no_h5py()
    test_load_sum()



"""
This module imports the algorithm classes to qclab.algorithms.
"""

from qclab.algorithms.mean_field import MeanField, MeanFieldAbInitio
from qclab.algorithms.fewest_switches_surface_hopping import (
    FewestSwitchesSurfaceHopping,
    FewestSwitchesSurfaceHoppingAbInitio,
)



"""
This module contains the Q-Chem interface for ab initio calculations.
"""

import os
import subprocess
import logging
import numpy as np
from qclab import numerical_constants

logger = logging.getLogger(__name__)


class QCLabQChemInterface:
    """
    Q-Chem interface for ab initio calculations.
    """

    def __init__(
        self,
        atom_positions,  # In units of Bohr.
        atom_masses,  # In units of electron mass.
        atom_names,
        label="qchem",
        folder_scratch="qclab_job",
        num_threads=1,
        num_procs=1,
        **kwargs,
    ):
        self.method_es = kwargs["method_es"]

        # Set default parameters.
        if kwargs.get("exchange", None) is None:
            kwargs["exchange"] = "B3LYP"
            self.method_es = "tddft"
        if kwargs.get("basis", None) is None:
            kwargs["basis"] = "6-31G*"
        # Ensuring that all kwargs are lowercase.
        kwargs = {str(k).lower(): v for k, v in kwargs.items()}
        self.method_es = self.method_es.lower()
        # Create atoms dict.
        atoms = {}
        for i, atom_name in enumerate(atom_names):
            atoms[i] = {
                "name": atom_name,
                "position": atom_positions[i] / numerical_constants.ANGSTROM_TO_BOHR,
                "mass": atom_masses[i] / numerical_constants.AMU_TO_EMASS,
            }
        # Store instance attributes.
        self.atom_names = atom_names
        self.atom_masses = atom_masses
        self.atom_positions = atom_positions
        self.atoms = atoms
        self.folder_scratch = folder_scratch
        self.label = label
        self.kwargs = kwargs
        self.num_threads = num_threads
        self.num_procs = num_procs
        self.results = {}
        # Properties implemented.
        self.implemented_properties = [
            "energy",
            "gradient",
            "derivative_coupling",
            "frequency",
            "wf_overlaps",
        ]
        # Job specifications for Q-Chem input file.
        self.job_templates = {
            "energy": {
                "name": "energy",
                "write_derivative_coupling": False,
                "excited_amplitudes": False,
                "qchem_parameters": {
                    "cis": {
                        "jobtype": "SP",
                        "cis_n_roots": self.kwargs.get("cis_n_roots"),
                        "cis_singlets": self.kwargs.get("cis_singlets"),
                        "cis_triplets": self.kwargs.get("cis_triplets"),
                        "basis": self.kwargs.get("basis"),
                        "exchange": self.kwargs.get("exchange"),
                        "sym_ignore": "TRUE",
                    },
                    "tddft": {
                        "jobtype": "SP",
                        "cis_n_roots": self.kwargs.get("cis_n_roots"),
                        "cis_singlets": self.kwargs.get("cis_singlets"),
                        "cis_triplets": self.kwargs.get("cis_triplets"),
                        "basis": self.kwargs.get("basis"),
                        "exchange": self.kwargs.get("exchange"),
                        "sym_ignore": "TRUE",
                    },
                },
            },
            "gradient": {
                "name": "gradient",
                "write_derivative_coupling": False,
                "qchem_parameters": {
                    "cis": {
                        "jobtype": "FORCE",
                        "basis": self.kwargs.get("basis"),
                        "exchange": self.kwargs.get("exchange"),
                        "sym_ignore": "TRUE",
                    },
                    "tddft": {
                        "jobtype": "FORCE",
                        "basis": self.kwargs.get("basis"),
                        "exchange": self.kwargs.get("exchange"),
                        "sym_ignore": "TRUE",
                    },
                },
            },
            "frequency": {
                "name": "frequency",
                "write_derivative_coupling": False,
                "qchem_parameters": {
                    "cis": {
                        "jobtype": "FREQ",
                        "basis": self.kwargs.get("basis"),
                        "exchange": self.kwargs.get("exchange"),
                        "sym_ignore": "TRUE",
                    },
                    "tddft": {
                        "jobtype": "FREQ",
                        "basis": self.kwargs.get("basis"),
                        "exchange": self.kwargs.get("exchange"),
                        "sym_ignore": "TRUE",
                    },
                },
            },
            "derivative_coupling": {
                "name": "derivative_coupling",
                "write_derivative_coupling": True,
                "qchem_parameters": {
                    "cis": {
                        "jobtype": "SP",
                        "basis": self.kwargs.get("basis"),
                        "method": self.kwargs.get("exchange"),
                        "cis_n_roots": self.kwargs.get("cis_n_roots"),
                        "cis_singlets": self.kwargs.get("cis_singlets"),
                        "cis_triplets": self.kwargs.get("cis_triplets"),
                        "calc_nac": "TRUE",
                        "cis_der_numstate": int(self.kwargs.get("cis_n_roots")) + 1,
                        "sym_ignore": "TRUE",
                    },
                    "tddft": {
                        "jobtype": "SP",
                        "basis": self.kwargs.get("basis"),
                        "method": self.kwargs.get("exchange"),
                        "cis_n_roots": self.kwargs.get("cis_n_roots"),
                        "cis_singlets": self.kwargs.get("cis_singlets"),
                        "cis_triplets": self.kwargs.get("cis_triplets"),
                        "calc_nac": "TRUE",
                        "cis_der_numstate": int(self.kwargs.get("cis_n_roots")) + 1,
                        "sym_ignore": "TRUE",
                    },
                },
            },
            "wf_overlaps": {
                "name": "wf_overlaps",
                "write_derivative_coupling": False,
                "qchem_parameters": {
                    "cis": {
                        "jobtype": "SP",
                        "basis": self.kwargs.get("basis"),
                        "exchange": self.kwargs.get("exchange"),
                        "sym_ignore": "TRUE",
                    },
                    "tddft": {
                        "jobtype": "SP",
                        "basis": self.kwargs.get("basis"),
                        "exchange": self.kwargs.get("exchange"),
                        "sym_ignore": "TRUE",
                    },
                },
            },
        }

    def execute(self):
        """
        Create and execute the command for the Q-Chem calculation.
        """
        # Build command.
        command = "qchem"
        if self.num_procs != 1:
            command += f" -np {self.num_procs}"
        if self.num_threads != 1:
            command += f" -nt {self.num_threads}"
        command += f" {self.label}.inp {self.label}.out"
        command += " " + self.folder_scratch
        command += " > /dev/null 2>&1"
        subprocess.run(command, shell=True, cwd=os.getcwd())

    def _build_job_specs(self, properties):
        """
        Build the job specifications for the requested properties.

        Each job spec is a dict with:
        name  : label (e.g. 'gradient', 'derivative_coupling', 'frequency', 'energy')
        write_derivative_coupling : bool, whether to emit $derivative_coupling
        qchem_parameters : dict of dicts, keyed by method_es, of Q-Chem parameters to be written in $rem section.
        """

        if properties is None:
            properties = ["energy"]
        properties = [prop.lower() for prop in properties]
        # Check if the asked properties are implemented.
        self._check_requested_properties(properties)
        properties = list(dict.fromkeys(properties))
        return [self.job_templates[p].copy() for p in properties]

    def _check_requested_properties(self, job_requested):
        """
        Checks if the requested properties are implemented.
        """
        for job in job_requested:
            if job not in self.implemented_properties:
                raise ValueError(f"Requested property '{job}' is not implemented.")

    def _compute_job_read_flag(self, num_jobs, indx):
        """
        Compute the flag for the $molecule section of each job in the input file.
        gradient/wf_overlaps: flag=0 for single-job input or first job,
        otherwise offset by +1
        """
        return 0 if (num_jobs == 1 or indx == 0) else (indx + 1)

    def _add_parameters_to_job_spec(self, job_spec):
        """
        Add parameters from kwargs to job_spec.
        """
        if self.kwargs.get(job_spec["name"], None) is not None:
            self.kwargs[job_spec["name"]] = {
                str(k).lower(): v for k, v in self.kwargs[job_spec["name"]].items()
            }
            job_spec["qchem_parameters"][self.method_es] = {
                str(k).lower(): v
                for k, v in job_spec["qchem_parameters"][self.method_es].items()
            }
            for key, value in self.kwargs[job_spec["name"]].items():
                if key in job_spec["qchem_parameters"][self.method_es]:
                    job_spec["qchem_parameters"][self.method_es][key] = value
                elif key not in job_spec["qchem_parameters"][self.method_es]:
                    job_spec["qchem_parameters"][self.method_es][key] = value
        return job_spec

    def _write_job(self, file_obj, job_spec, **kwargs):
        """
        Write  Q-Chem jobs (single job or multiple jobs that share same geometry).

        The number of jobs is determined by the number of properties requested and
        the writing is handled by the corresponding method for each property.
        """
        num_jobs = len(job_spec)
        for i, job in enumerate(job_spec):
            flag = self._compute_job_read_flag(num_jobs, i)
            job = self._add_parameters_to_job_spec(job)
            if "gradient" in job["name"]:
                self._write_gradient_jobs(
                    file_obj,
                    job,
                    flag,
                    kwargs["gradient"].get("state_inds_gradient", None),
                )
                continue
            elif "wf_overlaps" in job["name"]:
                if kwargs["wf_overlaps"]["atom_positions_previous"] is None:
                    raise ValueError(
                        "Previous geometry must be provided"
                        "when requesting wf_overlaps."
                    )
                atoms_previous = {}
                for i, atom_name in enumerate(self.atom_names):
                    atoms_previous[i] = {
                        "name": atom_name,
                        "position": kwargs["wf_overlaps"]["atom_positions_previous"][i]
                        / numerical_constants.ANGSTROM_TO_BOHR,
                        "mass": self.atom_masses[i] / numerical_constants.AMU_TO_EMASS,
                    }
                self._write_wf_overlaps_jobs(file_obj, job, atoms_previous, flag)
                continue
            else:
                self._write_comments(file_obj, job["name"], ind=i)
                self._write_molecule_section(file_obj, ind=i)
                if job["name"] == "energy" and kwargs["energy"]["excited_amplitudes"]:
                    job["qchem_parameters"][self.method_es]["GUI"] = "2"
                    job["qchem_parameters"][self.method_es]["IQMOL_FCHK"] = "TRUE"
                if job["name"] == "derivative_coupling":
                    self._get_state_inds_derivative_coupling(
                        job,
                        kwargs["derivative_coupling"].get(
                            "state_inds_derivative_coupling", None
                        ),
                    )
                    if len(self.state_inds_derivative_coupling) != int(
                        job["qchem_parameters"][self.method_es].get("cis_der_numstate")
                    ):
                        job["qchem_parameters"][self.method_es]["cis_der_numstate"] = (
                            len(self.state_inds_derivative_coupling)
                        )
                self._write_rem_section(
                    file_obj,
                    job,
                )
                if job["write_derivative_coupling"]:
                    self._write_derivative_coupling(
                        file_obj,
                        job,
                    )

    def _get_state_inds_gradient(self, state_inds_gradient):
        """
        Return a list of state indices for gradient calculations.

        If no indices are provided, defaults to all states (ground + excited).
        """
        if state_inds_gradient is None:
            num_states = int(self.kwargs.get("cis_n_roots")) + 1
            return [i for i in range(num_states)]
        elif isinstance(state_inds_gradient, (int, np.integer)):
            return [state_inds_gradient]
        else:
            return state_inds_gradient

    def _write_gradient_jobs(
        self, file_obj, job_spec, flag=0, state_inds_gradient=None
    ):
        """
        Handles the writing for gradient jobs requested.
        """
        self.state_inds_gradient = self._get_state_inds_gradient(state_inds_gradient)
        for j, i in enumerate(self.state_inds_gradient):
            if i > 0:
                job_spec["qchem_parameters"][self.method_es]["cis_state_deriv"] = str(i)
                job_spec["qchem_parameters"][self.method_es]["cis_n_roots"] = (
                    self.kwargs.get("cis_n_roots")
                )
                job_spec["qchem_parameters"][self.method_es]["cis_singlets"] = (
                    self.kwargs.get("cis_singlets")
                )
                job_spec["qchem_parameters"][self.method_es]["cis_triplets"] = (
                    self.kwargs.get("cis_triplets")
                )
            if flag == 0:
                ind = j
            else:
                ind = flag
            self._write_comments(file_obj, f"gradient state {i}", ind)
            self._write_molecule_section(file_obj, ind)
            self._write_rem_section(file_obj, job_spec)

    def _write_wf_overlaps_jobs(self, file_obj, job_spec, atoms_previous, flag=0):
        """
        Handles the writing for wavefunction overlaps jobs requested.
        """
        ind = flag
        # Create the input for the previous geometry.
        self._write_comments(file_obj, job_spec["name"] + " previous step", ind)
        # Write geometry.
        file_obj.write("$molecule\n")
        charge, mult = self._get_charge_and_multiplicity()
        file_obj.write("   %d %d\n" % (charge, mult))
        self._write_geometry(file_obj, atoms_previous)
        file_obj.write("$end\n\n")
        # Job definition.
        keys_for_wf_overlaps = ["mo_overlaps_two_geoms"]
        if all(
            key in job_spec["qchem_parameters"][self.method_es]
            for key in keys_for_wf_overlaps
        ):
            pass
        else:
            job_spec["qchem_parameters"][self.method_es]["mo_overlaps_two_geoms"] = 1
        if (
            job_spec["qchem_parameters"][self.method_es].get("mo_overlaps_two_geoms")
            != 1
        ):
            job_spec["qchem_parameters"][self.method_es]["mo_overlaps_two_geoms"] = 1
        self._write_rem_section(file_obj, job_spec)

        # Creating input for the current geometry.
        self._write_comments(file_obj, job_spec["name"] + " current step", ind + 1)
        self._write_molecule_section(file_obj, ind=0)
        job_spec["qchem_parameters"][self.method_es]["mo_overlaps_two_geoms"] = 2
        self._write_rem_section(file_obj, job_spec)

    def _write_comments(self, file_obj, job, ind=0):
        """
        Write comments and job separator for Q-Chem input file.
        """
        if ind != 0:
            file_obj.write("@@@\n\n")
        file_obj.write("$comment\n")
        file_obj.write(f"QC Lab generated input file ({job} job)\n")
        file_obj.write("$end\n\n")

    def _get_charge_and_multiplicity(self):
        """
        Return the molecular charge and spin multiplicity from kwargs
        if provided, otherwise raise an error.
        """
        charge = self.kwargs.get("charge", None)
        multiplicity = self.kwargs.get("multiplicity", None)
        if multiplicity is None or charge is None:
            raise ValueError("Charge and multiplicity must be provided.")
        else:
            multiplicity = int(multiplicity)
            charge = int(charge)
        return charge, multiplicity

    def _write_geometry(self, file_obj, geometry):
        """
        Write atomic coordinates to the Q-Chem input file.
        """
        for i in range(len(self.atom_names)):
            file_obj.write(
                "   {}  {:f}  {:f}  {:f}\n".format(
                    geometry[i]["name"],
                    geometry[i]["position"][0],
                    geometry[i]["position"][1],
                    geometry[i]["position"][2],
                )
            )

    def _write_molecule_section(self, file_obj, ind=0):
        """
        Write the $molecule section to the Q-Chem input file.

        For the first job (ind=0), writes charge, multiplicity, and geometry.
        For subsequent jobs, writes 'read' to reuse the previous geometry.
        """
        file_obj.write("$molecule\n")
        if ind == 0:
            charge, multiplicity = self._get_charge_and_multiplicity()
            file_obj.write("   %d %d\n" % (charge, multiplicity))
            self._write_geometry(file_obj, self.atoms)
        else:
            file_obj.write("   read\n")
        file_obj.write("$end\n\n")

    def _write_rem_section(self, file_obj, job_spec):
        """
        Write the $rem section to the Q-Chem input file from the job spec parameters.
        """
        file_obj.write("$rem\n")
        for parameter, v_str in job_spec["qchem_parameters"][self.method_es].items():
            if isinstance(v_str, str):
                v_str = v_str.upper()
            if v_str == "NONE" or v_str is None:
                continue
            file_obj.write("   %-25s   %s\n" % (parameter.upper(), v_str))
        file_obj.write("$end\n\n")

    def _get_state_inds_derivative_coupling(
        self, job, state_inds_derivative_coupling=None
    ):
        """
        Return a list/tuple of state indices for derivative coupling calculations.

        If no indices are provided, defaults to all states (ground + excited) as specified
        by cis_der_numstate (cis_n_roots+1) in the input file.
        """
        if state_inds_derivative_coupling is None:
            n_s = int(job["qchem_parameters"][self.method_es].get("cis_der_numstate"))
            state_inds_derivative_coupling = [i for i in range(n_s)]
        elif isinstance(state_inds_derivative_coupling, (int, np.integer)):
            state_inds_derivative_coupling = [
                i for i in range(state_inds_derivative_coupling + 1)
            ]
        self.state_inds_derivative_coupling = state_inds_derivative_coupling

    def _write_derivative_coupling(self, file_obj, job):
        """
        Write the $derivative_coupling section to the Q-Chem input file.
        """
        file_obj.write("$derivative_coupling\n")
        file_obj.write(
            f"{self.state_inds_derivative_coupling[0]} is the reference state\n"
        )
        file_obj.write(
            " ".join(str(s) for s in self.state_inds_derivative_coupling) + "\n"
        )
        file_obj.write("$end\n\n")

    def read_results(self, **kwargs):
        """
        Read and parse the Q-Chem output file for the requested properties.

        Results are stored in ``self.results``.

        .. rubric:: Args
        **kwargs:
            Keyword arguments keyed by property name (e.g. 'energy', 'gradient').
            Each value is a dict of options specific to that property.
        """
        properties = kwargs.keys()
        if properties is None:
            properties = ["energy"]
        properties = [prop.lower() for prop in properties]
        num_atoms = len(self.atoms)
        filename = self.label + ".out"
        with open(filename, "r") as file_obj:
            file_content = file_obj.read()

        self._check_normal_termination(file_content, properties)
        self._organize_data_file(file_content, properties, num_atoms, **kwargs)

    def _get_number_of_jobs(self, properties):
        """
        Return the total number of Q-Chem jobs for the requested properties.

        Gradient and wavefunction overlap calculations may require
        multiple jobs beyond one per property.
        """
        num_jobs = len(properties)
        if "gradient" in properties:
            num_jobs += len(self.state_inds_gradient) - 1
        if "wf_overlaps" in properties:
            num_jobs += 1
        return num_jobs

    def _check_normal_termination(self, file_content, properties):
        """
        Check Q-Chem normal termination for the number of jobs
        corresponding to the requested properties, if the number of jobs
        is not equal to the  number of normal termination flags in
        the output file, raise an error.
        """
        normal_termination = "for using Q-Chem"
        num_jobs = self._get_number_of_jobs(properties)
        if file_content.count(normal_termination) != num_jobs:
            raise ValueError("Q-Chem did not terminate normally.")

    def _split_file_by_job(self, file_content, properties, **kwargs):
        """
        Split the Q-Chem output file content into sections for each job.
        """
        num_jobs = self._get_number_of_jobs(properties)
        splitting_flag = "Running Job"
        if num_jobs == 1:
            return [file_content]
        else:
            return file_content.split(splitting_flag)[1 : num_jobs + 1]

    def _organize_data_file(self, file_content, properties, num_atoms, **kwargs):
        """
        Organize Q-Chem data according to the requested properties.
        """
        split_files = self._split_file_by_job(file_content, properties, **kwargs)
        i = 0
        for job in properties:
            if "gradient" in job:
                gradient_files = split_files[i : i + len(self.state_inds_gradient)]
                i += len(self.state_inds_gradient)
                self._pull_property(
                    job,
                    gradient_files,
                    num_atoms,
                )
            elif "wf_overlaps" in job:
                overlaps_files = split_files[i : i + 1]
                i += 2
                self._pull_property(
                    job,
                    overlaps_files,
                    num_atoms,
                    amplitudes_previous=kwargs["wf_overlaps"].get(
                        "amplitudes_previous", None
                    ),
                    amplitudes_current=kwargs["wf_overlaps"].get(
                        "amplitudes_current", None
                    ),
                )
            else:
                job_file = split_files[i]
                i += 1
                excited_amplitudes = (
                    kwargs["energy"].get("excited_amplitudes", False)
                    if "energy" in job
                    else False
                )
                self._pull_property(
                    job,
                    job_file.splitlines(),
                    num_atoms,
                    excited_amplitudes=excited_amplitudes,
                )

    def _pull_property(self, property, file_obj, num_atoms, **kwargs):
        """
        Pull the requested property from the Q-Chem output file section.

        Determines which parsing method to call based on the property type and
        stores the results in ``self.results``.
        """
        if "energy" in property:
            self._pull_energy(
                file_obj, excited_amplitudes=kwargs.get("excited_amplitudes", False)
            )
        elif "gradient" in property:
            self._pull_gradient(file_obj, num_atoms)
        elif "frequency" in property:
            self._pull_vibration(file_obj, num_atoms)
        elif "derivative_coupling" in property:
            self._pull_derivative_coupling(file_obj, num_atoms)
        elif "wf_overlaps" in property:
            self._pull_overlaps(
                amplitudes_previous=kwargs.get("amplitudes_previous", None),
                amplitudes_current=kwargs.get("amplitudes_current", None),
            )
        else:
            raise ValueError("This type of calculation has not been implemented yet.")

    def _pull_vibration(self, file_obj, num_atoms):
        """
        Parse vibrational frequencies and normal modes from Q-Chem output.

        Stores 'frequency' and 'normal_mode' in ``self.results``.
        """
        ind_modes = []
        for i, line in enumerate(file_obj):
            if "Mode:" in line:
                ind_modes.append(i)
        freqs = []
        for i in ind_modes:
            freqs.append(file_obj[i + 1].split()[1:])
        freqs = np.array(freqs, dtype=float).flatten()
        modes = np.zeros((len(freqs), num_atoms, 3))
        i_count = 0
        # Number of lines between the Mode: line and the displacement-matrix line in a Q-Chem output file.
        num_lines = 8
        for i in ind_modes:
            for j in range(num_atoms):
                temporal_data = file_obj[i + num_lines + j].split()[1:]
                num_modes = int(len(temporal_data) / 3)
                for k in range(num_modes):
                    modes[i_count + k, j, :] = temporal_data[k * 3 : k * 3 + 3]
            i_count += num_modes

        self.results["frequency"] = freqs
        self.results["normal_mode"] = modes

    def _pull_energy(self, file_obj, excited_amplitudes=False):
        """
        Parse ground and excited state energies from Q-Chem output.

        Optionally extracts excited state amplitudes from the FCHK file.
        Stores 'energy' (and 'excited_amplitudes' if requested) in ``self.results``.
        """
        # Get the number of states to be extracted from the output file.
        if self.kwargs.get("cis_n_roots") is None:
            nt_states = 1
        else:
            nt_states = int(self.kwargs.get("cis_n_roots")) + 1

        keywords_for_gs_energy = [
            "SCF time:",
            "Timing for Total SCF",
            "Total energy in the final basis set",
        ]
        keywords_for_excited_energy = ["Total energy for state"]

        i = 0
        energy = np.zeros(nt_states)
        for j, line in enumerate(file_obj):
            # Extract ground state energy.
            if any(word in line for word in keywords_for_gs_energy):
                try:
                    energy[i] = float(file_obj[j + 2].split()[-1])
                    i += 1
                except:
                    pass
            # Extract excited state energies.
            elif any(word in line for word in keywords_for_excited_energy):
                if i >= nt_states:
                    logger.warning(
                        "More states found in output file than specified by cis_n_roots."
                        " Only the first %d states will be extracted." % nt_states
                    )
                    break
                energy[i] = float(line.split()[-2])
                i += 1
        self.results["energy"] = energy
        # Extract excited state amplitudes if requested.
        if excited_amplitudes:
            self.results["excited_amplitudes"] = {}
            (
                x_alpha,
                y_alpha,
                num_basis_functions,
                num_alpha_electrons,
                num_excited_states,
            ) = self._pull_excited_amplitudes()
            self.results["excited_amplitudes"]["x"] = x_alpha
            self.results["excited_amplitudes"]["y"] = y_alpha
            self.results["excited_amplitudes"][
                "num_basis_functions"
            ] = num_basis_functions
            self.results["excited_amplitudes"][
                "num_alpha_electrons"
            ] = num_alpha_electrons
            self.results["excited_amplitudes"][
                "num_excited_states"
            ] = num_excited_states

    def _pull_gradient(self, file_obj, num_atoms):
        """
        Parse gradients for each state from Q-Chem output.

        Stores 'gradient' in ``self.results``.
        """
        gradient = np.zeros((num_atoms, 3, len(self.state_inds_gradient)))
        flag_words_grad = [
            "Gradient of the state energy (including CIS Excitation Energy)",
            "Gradient of SCF Energy",
        ]
        flag_words_out_grad = ["Gradient time", "Max gradient component"]
        for state_ind in range(len(self.state_inds_gradient)):
            flag_in = next(
                flag for flag in flag_words_grad if flag in file_obj[state_ind]
            )
            flag_out = next(
                flag for flag in flag_words_out_grad if flag in file_obj[state_ind]
            )
            gradient_matrix = file_obj[state_ind].split(flag_in)[1]
            gradient_matrix = gradient_matrix.split(flag_out)[0]
            # Ensure that the gradient matrix is clean.
            for f in flag_words_out_grad:
                if gradient_matrix.count(f) != 0:
                    gradient_matrix = gradient_matrix.split(f)[0]
            stop_ind = 0
            i_count = 1
            gradient_matrix = gradient_matrix.splitlines()
            while stop_ind < num_atoms:
                jndx = np.array(gradient_matrix[i_count].split(), dtype=int) - 1
                gradient[jndx, 0, state_ind] = np.array(
                    gradient_matrix[i_count + 1].split()[1:], dtype=float
                )
                gradient[jndx, 1, state_ind] = np.array(
                    gradient_matrix[i_count + 2].split()[1:], dtype=float
                )
                gradient[jndx, 2, state_ind] = np.array(
                    gradient_matrix[i_count + 3].split()[1:], dtype=float
                )
                i_count += 4
                stop_ind = jndx.max() + 1
        self.results["gradient"] = gradient

    def _pull_derivative_coupling(self, file_obj, num_atoms, etf=True):
        """
        Parse derivative couplings between electronic states from Q-Chem output.

        Stores 'derivative_coupling' in ``self.results`` as a dictionary
        keyed by state-pair tuples.
        """
        l_states = []  # Line that identifies the states involved in derivative coupling
        l_etf = (
            []
        )  # Line that identifies the location of derivative coupling's values with ETF corrections.
        l_no_etf = (
            []
        )  # Line that identifies the location of derivative coupling's values without ETF corrections.
        for i, line in enumerate(file_obj):
            if "between states" in line:
                l_states.append(i)
            elif "with ETF" in line:
                l_etf.append(i)
            elif "without ETF" in line:
                l_no_etf.append(i)

        # Gather derivative coupling values.
        derivative_coupling_matrix = np.zeros((num_atoms, 3))
        derivative_coupling_dictionary = {}
        for i, states in enumerate(l_states):
            temp = file_obj[states].split()
            # States involved in derivative coupling.
            i_st = int(temp[-3])
            j_st = int(temp[-1])
            # Base line for derivative coupling values.
            base = l_etf[i] if etf else l_no_etf[i]
            for j in range(num_atoms):
                derivative_coupling_matrix[j, :] = np.array(
                    file_obj[base + 3 + j].split()[1:], dtype=float
                )
            derivative_coupling_dictionary[(i_st, j_st)] = (
                derivative_coupling_matrix.copy()
            )
        self.results["derivative_coupling"] = derivative_coupling_dictionary

    def _pull_overlaps(self, amplitudes_previous=None, amplitudes_current=None):
        """
        Parse wavefunction overlaps between two geometries from Q-Chem output.

        Determine which method to use for computing overlaps based on the method
        used for excited state calculations (TDDFT or CIS) and call the corresponding method.
        The overlaps are stored in ``self.results["wf_overlaps"]`` as a matrix.
        """
        if amplitudes_previous is None:
            raise ValueError(
                "Previous excited state amplitudes must be provided \n"
                "to compute wavefunction overlaps."
            )
        if amplitudes_current is None:
            raise ValueError(
                "Current excited state amplitudes must be provided \n"
                "to compute wavefunction overlaps."
            )
        # Create instance attributes needed for computing overlaps.
        self.num_basis_functions = amplitudes_current["num_basis_functions"]
        self.num_alpha_electrons = amplitudes_current["num_alpha_electrons"]
        self.num_excited_states = amplitudes_current["num_excited_states"]
        # Extract molecular orbitals overlaps between two geometries.
        mo_overlaps = self._pull_mo_overlaps()
        # Compute overlaps matrices.
        if self.method_es.lower() == "tddft":
            self._get_overlaps_tddft(
                amplitudes_previous["x"],
                amplitudes_previous["y"],
                amplitudes_current["x"],
                amplitudes_current["y"],
                mo_overlaps,
            )
        elif self.method_es.lower() == "cis":
            self._get_overlaps_cis(
                amplitudes_previous["x"],
                amplitudes_current["x"],
                mo_overlaps,
            )
        else:
            raise ValueError(
                f"Method {self.method_es} not recognized. Supported methods are 'tddft' and 'cis'."
            )

    def _get_overlaps_cis(self, alpha_prev, alpha_curr, mo_overlaps):
        """
        Compute wavefunction overlaps using the CIS formalism.

        The results are stored in ``self.results["wf_overlaps"]`` as a matrix.
        """
        s_oo = mo_overlaps[0 : self.num_alpha_electrons, 0 : self.num_alpha_electrons]
        s_ov = mo_overlaps[0 : self.num_alpha_electrons, self.num_alpha_electrons :]
        s_vo = mo_overlaps[self.num_alpha_electrons :, 0 : self.num_alpha_electrons]
        s_vv = mo_overlaps[self.num_alpha_electrons :, self.num_alpha_electrons :]
        # Compute <GS^(1)|GS^(2)>.
        sing, log_determinant = np.linalg.slogdet(s_oo)
        gs_overlap = sing * np.exp(log_determinant)
        # Compute <GS^(1)|ES_i^(2)>.
        overlaps_gs_ex = self._compute_overlaps_gs_ex(alpha_curr, s_oo, s_ov)
        # Compute <ES_i^(1)|GS^(2)>.
        overlaps_ex_gs = self._compute_overlaps_ex_gs(alpha_prev, s_oo, s_vo)
        # Compute <ES_i^(1)|ES_j^(2)>.
        overlaps_ex_ex = self._compute_overlaps_ex_ex(
            alpha_prev, alpha_curr, s_oo, s_ov, s_vo, s_vv
        )
        self.results["wf_overlaps"] = np.zeros(
            (self.num_excited_states + 1, self.num_excited_states + 1)
        )
        self.results["wf_overlaps"][0, 0] = gs_overlap
        self.results["wf_overlaps"][0, 1:] = 2.0 * overlaps_gs_ex
        self.results["wf_overlaps"][1:, 0] = 2.0 * overlaps_ex_gs
        self.results["wf_overlaps"][1:, 1:] = 2.0 * overlaps_ex_ex

    def _get_overlaps_tddft(self, x_prev, y_prev, x_curr, y_curr, mo_overlaps):
        """
        Compute wavefunction overlaps using the TDDFT formalism.

        The results are stored in ``self.results["wf_overlaps"]`` as a matrix.
        """
        s_oo = mo_overlaps[0 : self.num_alpha_electrons, 0 : self.num_alpha_electrons]
        s_ov = mo_overlaps[0 : self.num_alpha_electrons, self.num_alpha_electrons :]
        s_vo = mo_overlaps[self.num_alpha_electrons :, 0 : self.num_alpha_electrons]
        s_vv = mo_overlaps[self.num_alpha_electrons :, self.num_alpha_electrons :]
        # Compute <GS^(1)|GS^(2)>.
        sing, log_determinant = np.linalg.slogdet(s_oo)
        gs_overlap = sing * np.exp(log_determinant)
        # Compute <GS^(1)|ES_i^(2)>.
        overlaps_gs_ex_x = self._compute_overlaps_gs_ex(x_curr, s_oo, s_ov)
        overlaps_gs_ex_y = self._compute_overlaps_gs_ex(y_curr, s_oo, s_ov)
        overlaps_gs_ex = overlaps_gs_ex_x - overlaps_gs_ex_y
        # Compute <ES_i^(1)|GS^(2)>.
        overlaps_ex_gs_x = self._compute_overlaps_ex_gs(x_prev, s_oo, s_vo)
        overlaps_ex_gs_y = self._compute_overlaps_ex_gs(y_prev, s_oo, s_vo)
        overlaps_ex_gs = overlaps_ex_gs_x - overlaps_ex_gs_y
        # Compute <ES_i^(1)|ES_j^(2)>.
        overlaps_ex_ex_x_x = self._compute_overlaps_ex_ex(
            x_prev, x_curr, s_oo, s_ov, s_vo, s_vv
        )
        overlaps_ex_ex_x_y = self._compute_overlaps_ex_ex(
            x_prev, y_curr, s_oo, s_ov, s_vo, s_vv
        )
        overlaps_ex_ex_y_x = self._compute_overlaps_ex_ex(
            y_prev, x_curr, s_oo, s_ov, s_vo, s_vv
        )
        overlaps_ex_ex_y_y = self._compute_overlaps_ex_ex(
            y_prev, y_curr, s_oo, s_ov, s_vo, s_vv
        )
        overlaps_ex_ex = (
            overlaps_ex_ex_x_x
            - overlaps_ex_ex_x_y
            - overlaps_ex_ex_y_x
            + overlaps_ex_ex_y_y
        )
        self.results["wf_overlaps"] = np.zeros(
            (self.num_excited_states + 1, self.num_excited_states + 1)
        )
        self.results["wf_overlaps"][0, 0] = gs_overlap
        self.results["wf_overlaps"][0, 1:] = 2.0 * overlaps_gs_ex
        self.results["wf_overlaps"][1:, 0] = 2.0 * overlaps_ex_gs
        self.results["wf_overlaps"][1:, 1:] = 2.0 * overlaps_ex_ex

    def _compute_overlaps_gs_ex(self, excited_amplitudes, s_oo, s_ov):
        """
        Compute overlap between ground state at geometry 1 and excited states at geometry 2.
        """
        a_matrix = s_oo
        sign, log_determinant = np.linalg.slogdet(a_matrix)
        determinant_a_matrix = sign * np.exp(log_determinant)
        w_matrix = np.linalg.solve(a_matrix, s_ov)
        x = np.asarray(excited_amplitudes)
        overlaps_gs_ex = determinant_a_matrix * np.einsum("eia,ia->e", x, w_matrix)
        return overlaps_gs_ex

    def _compute_overlaps_ex_gs(self, excited_amplitudes, s_oo, s_vo):
        """
        Compute overlap between excited states at geometry 1 and ground state at geometry 2.
        """
        a_matrix = s_oo
        sign, log_determinant = np.linalg.slogdet(a_matrix)
        determinant_a_matrix = sign * np.exp(log_determinant)
        w_matrix = np.linalg.solve(a_matrix.T, s_vo.T).T
        x = np.asarray(excited_amplitudes)
        overlaps_ex_gs = determinant_a_matrix * np.einsum("eia,ai->e", x, w_matrix)
        return overlaps_ex_gs

    def _compute_overlaps_ex_ex(
        self, geometry_1_amplitudes, geometry_2_amplitudes, s_oo, s_ov, s_vo, s_vv
    ):
        """
        Compute overlap between excited states at geometry 1 and excited states at geometry 2.
        """
        a_matrix = s_oo
        num_occupied_orbitals = a_matrix.shape[0]
        x = np.asarray(geometry_1_amplitudes)
        x2 = np.asarray(geometry_2_amplitudes)
        sign, log_determinant = np.linalg.slogdet(a_matrix)
        determinant_a_matrix = sign * np.exp(log_determinant)
        inverse_a_matrix = np.linalg.solve(a_matrix, np.eye(num_occupied_orbitals))
        g_matrix = inverse_a_matrix @ s_ov
        overlap = np.zeros((x.shape[0], x2.shape[0]))
        for j in range(num_occupied_orbitals):
            aj = a_matrix[j, :]
            column_j = inverse_a_matrix[:, j]
            delta = s_vo - aj[None, :]
            delta_dot_column_j = delta @ column_j
            delta_dot_g = delta @ g_matrix
            for i in range(num_occupied_orbitals):
                q = inverse_a_matrix[i, j]
                gi = g_matrix[i, :]
                aji = a_matrix[j, i]
                delta_i = delta[:, i]
                alpha = s_vv - s_vo[:, i][:, None] - s_ov[j, :][None, :] + aji
                b21 = (delta_dot_g - delta_i[:, None]) + alpha * (gi[None, :] - 1.0)
                b22 = 1.0 + delta_dot_column_j[:, None] + alpha * q
                determinant_m = determinant_a_matrix * (gi[None, :] * b22 - q * b21)
                conjx = np.conj(x[:, j, :])
                x2ia = x2[:, i, :]
                overlap += np.einsum(
                    "eb,fa,ba->ef", conjx, x2ia, determinant_m, optimize=True
                )
        return overlap

    def _pull_mo_overlaps(self):
        """
        Read the molecular orbital overlaps between two geometries from the Q-Chem scratch folder.
        """
        qcscratch = os.environ["QCSCRATCH"]  # Q-Chem scratch folder.
        file_mo_overlap = os.path.join(
            qcscratch, self.folder_scratch, "MO-overlaps", "MO-overlap-TwoGeoms.txt"
        )
        with open(file_mo_overlap, "r") as file:
            data = file.readlines()
        overlaps_init = []
        for line in data:
            temp_values = np.array(line.split(), dtype=float)
            if len(temp_values) == 8:
                overlaps_init.append(temp_values[:-1])
            elif len(temp_values) == 0:
                continue
            else:
                overlaps_init.append(temp_values)
        length = len(np.concatenate(overlaps_init))
        if length != (self.num_basis_functions * self.num_basis_functions):
            raise ValueError("The number of overlap integrals read is not correct")
        mo_overlaps = np.zeros((self.num_basis_functions, self.num_basis_functions))
        num_blocks = int(len(overlaps_init) / self.num_basis_functions)
        for i in range(self.num_basis_functions):
            temp_values = np.concatenate(
                [
                    overlaps_init[self.num_basis_functions * j + i]
                    for j in range(num_blocks)
                ]
            )
            mo_overlaps[i, :] = temp_values
        mo_overlaps = mo_overlaps.T
        return mo_overlaps

    def _pull_excited_amplitudes(self):
        """
        Read the excited state amplitudes from the FCHK file generated by Q-Chem.

        Extracts the information of the excited state amplitudes for TDDFT or CIS calculations.
        It only supports closed-shell calculations.
        The information that is extracted includes the x and y amplitudes for TDDFT or
        alpha amplitudes for CIS, as well as the number of basis functions,
        the number of alpha electrons, and the number of excited states.

        The results are returned as a tuple of
        (x_alpha, y_alpha, num_basis_functions, num_alpha_electrons, num_excited_states).
        """
        file_fchk = self.label + ".fchk"
        with open(file_fchk, "r") as f:
            data = f.readlines()
        keywords_to_find = [
            "Number of alpha electrons",
            "Number of beta electrons",
            "Number of basis functions",
            "Number of Excited States",
            "Alpha X Amplitudes",
            "Alpha Y Amplitudes",
            "Alpha Amplitudes",
        ]
        keyword_to_stop = ["Beta X Amplitudes", "Beta Y Amplitudes", "Beta Amplitudes"]

        for i, line in enumerate(data):
            if keywords_to_find[0] in line:
                num_alpha_electrons = int(line.split()[-1])
            if keywords_to_find[1] in line:
                num_beta_electrons = int(line.split()[-1])
            if keywords_to_find[2] in line:
                num_basis_functions = int(line.split()[-1])
            if keywords_to_find[3] in line:
                num_excited_states = int(line.split()[-1])
                index_restart = i

        if num_alpha_electrons != num_beta_electrons:
            raise ValueError("The interface only supports closed-shell calculations.")
        x_alpha = []
        y_alpha = []
        for i, line in enumerate(data[index_restart:]):
            if keywords_to_find[4] in line:
                dimension_amplitudes = int(line.split()[-1])
                j = i + 1 + index_restart
                while keyword_to_stop[0] not in data[j]:
                    temp_values = np.array(data[j].split(), dtype=float)
                    x_alpha.append(temp_values)
                    j = j + 1
            if keywords_to_find[5] in line:
                j = i + 1 + index_restart
                while keyword_to_stop[1] not in data[j]:
                    temp_values = np.array(data[j].split(), dtype=float)
                    y_alpha.append(temp_values)
                    j = j + 1
            if keywords_to_find[6] in line:
                dimension_amplitudes = int(line.split()[-1])
                j = i + 1 + index_restart
                while keyword_to_stop[2] not in data[j]:
                    temp_values = np.array(data[j].split(), dtype=float)
                    x_alpha.append(temp_values)
                    y_alpha.append(np.zeros_like(temp_values))
                    j = j + 1

        # Check that the number of basis functions used for Q-Chem corresponds
        # to the number of basis functions determined from the basis set information.
        # Q-Chem sometimes automatically projects out near-linear dependencies,
        # which reduces the number of MOs below the number of basis functions.

        num_basis_functions_check = (
            int((dimension_amplitudes / num_excited_states) / num_alpha_electrons)
            + num_alpha_electrons
        )
        if num_basis_functions_check != num_basis_functions:
            logger.critical(
                "The number of basis functions used for Q-Chem does not correspond to "
                "the number of basis functions determined from the basis set information."
            )
            logger.info(
                "Q-Chem sometimes automatically projects out near-linear dependencies",
                "which reduces the number of MOs below the number of basis functions.",
            )
            raise ValueError(
                "The number of basis functions used for Q-Chem does not correspond to "
                "the number of basis functions determined from the basis set information."
            )

        x_alpha = np.concatenate(x_alpha)
        x_alpha = np.reshape(
            x_alpha,
            (
                num_excited_states,
                num_alpha_electrons,
                (num_basis_functions - num_alpha_electrons),
            ),
        )
        y_alpha = np.concatenate(y_alpha)
        y_alpha = np.reshape(
            y_alpha,
            (
                num_excited_states,
                num_alpha_electrons,
                (num_basis_functions - num_alpha_electrons),
            ),
        )
        return (
            x_alpha,
            y_alpha,
            num_basis_functions,
            num_alpha_electrons,
            num_excited_states,
        )

    def write_input(self, **kwargs):
        """
        Write the Q-Chem input file for the requested properties.
        """
        properties = kwargs.keys()
        filename = self.label + ".inp"
        job_specs = self._build_job_specs(properties)
        with open(filename, "w") as file_obj:
            self._write_job(file_obj, job_specs, **kwargs)



"""
This module contains the Model class for Tully's second problem, a dual avoided crossing.
"""

import numpy as np
from qclab.model import Model
from qclab import ingredients
from qclab import functions


class TullyProblemTwo(Model):
    """
    A Model class for Tully's second problem: a dual avoided crossing.

    Reference publication:
    Tully. J. Chem. Phys. 1990, 93 (2), 1061–1071. https://doi.org/10.1063/1.459170.
    """

    def __init__(self, constants=None):
        if constants is None:
            constants = {}
        self.default_constants = {
            "init_momentum": 10.0,
            "init_position": -25.0,
            "mass": 2000.0,
            "A": 0.1,
            "B": 0.28,
            "C": 0.015,
            "D": 0.06,
            "E_0": 0.05,
        }
        super().__init__(self.default_constants, constants)
        self.update_dh_qc_dzc = True
        self.update_h_q = False

    def _init_model(self, parameters, **kwargs):
        self.constants.num_quantum_states = 2
        self.constants.num_classical_coordinates = 1
        self.constants.classical_coordinate_mass = np.array(
            [self.constants.get("mass")]
        )
        self.constants.classical_coordinate_weight = np.array([1.0])
        self.constants.init_position = np.array([self.constants.get("init_position")])
        self.constants.init_momentum = np.array([self.constants.get("init_momentum")])
        return

    def h_qc(self, parameters, **kwargs):
        """
        Quantum-classical Hamiltonian for Tully's second problem.
        """
        z = kwargs["z"]
        batch_size = len(z)
        num_quantum_states = self.constants.num_quantum_states
        A = self.constants.get("A")
        B = self.constants.get("B")
        C = self.constants.get("C")
        D = self.constants.get("D")
        E_0 = self.constants.get("E_0")
        # Calculate q.
        m = self.constants.classical_coordinate_mass[np.newaxis, :]
        h = self.constants.classical_coordinate_weight[np.newaxis, :]
        q = functions.z_to_q(z, m, h)[:, 0]
        # Calculate matrix elements.
        v_12 = C * (np.exp(-D * (q**2)))
        v_22 = -A * (np.exp(-B * (q**2))) + E_0
        # Assemble Hamiltonian.
        h_qc = np.zeros(
            (batch_size, num_quantum_states, num_quantum_states), dtype=complex
        )
        h_qc[:, 0, 1] = v_12
        h_qc[:, 1, 0] = v_12
        h_qc[:, 1, 1] = v_22
        return h_qc

    def dh_qc_dzc(self, parameters, **kwargs):
        """
        Gradient w.r.t. to the conjugate z coordinate of the quantum-classical Hamiltonian
        for Tully's second problem.
        """
        z = kwargs["z"]
        batch_size = len(z)
        num_quantum_states = self.constants.num_quantum_states
        num_classical_coordinates = self.constants.num_classical_coordinates
        A = self.constants.get("A")
        B = self.constants.get("B")
        C = self.constants.get("C")
        D = self.constants.get("D")
        # Calculate q.
        m = self.constants.classical_coordinate_mass[np.newaxis, :]
        h = self.constants.classical_coordinate_weight[np.newaxis, :]
        q = functions.z_to_q(z, m, h)[:, 0]
        # Calculate phase-space gradients.
        dv_12_dq = -2 * C * D * q * np.exp(-D * q**2)
        dv_22_dq = 2 * A * B * q * np.exp(-B * q**2)
        # Convert to complex gradients.
        dv_12_dzc = functions.dqdp_to_dzc(dv_12_dq, None, m[0], h[0])
        dv_22_dzc = functions.dqdp_to_dzc(dv_22_dq, None, m[0], h[0])
        # Assemble indices.
        batch_idx = np.repeat(np.arange(batch_size, dtype=int), 3)
        coord_idx = np.zeros(3 * batch_size, dtype=int)
        state_i_idx = np.tile(np.array([0, 1, 1], dtype=int), batch_size)
        state_j_idx = np.tile(np.array([1, 0, 1], dtype=int), batch_size)
        inds = (batch_idx, coord_idx, state_i_idx, state_j_idx)
        # Assemble matrix elements.
        mels = np.empty(3 * batch_size, dtype=complex)
        mels[0::3] = dv_12_dzc
        mels[1::3] = dv_12_dzc
        mels[2::3] = dv_22_dzc
        # Assemble shape.
        shape = (
            batch_size,
            num_classical_coordinates,
            num_quantum_states,
            num_quantum_states,
        )
        return inds, mels, shape

    ingredients = [
        ("h_q", ingredients.h_q_two_level),
        ("h_qc", h_qc),
        ("h_c", ingredients.h_c_free),
        ("dh_qc_dzc", dh_qc_dzc),
        ("dh_c_dzc", ingredients.dh_c_dzc_free),
        ("init_classical", ingredients.init_classical_definite_position_momentum),
        ("hop", ingredients.hop_free),
        ("_init_model", _init_model),
    ]



"""
This module contains the Model class for Tully's third problem,
an extended coupling with reflection.
"""

import numpy as np
from qclab.model import Model
from qclab import ingredients
from qclab import functions


class TullyProblemThree(Model):
    """
    A Model class for Tully's third problem: an extended coupling with reflection.

    Reference publication:
    Tully. J. Chem. Phys. 1990, 93 (2), 1061–1071. https://doi.org/10.1063/1.459170.
    """

    def __init__(self, constants=None):
        if constants is None:
            constants = {}
        self.default_constants = {
            "init_momentum": 10.0,
            "init_position": -25.0,
            "mass": 2000.0,
            "A": 0.0006,
            "B": 0.1,
            "C": 0.9,
        }
        super().__init__(self.default_constants, constants)
        self.update_dh_qc_dzc = True
        self.update_h_q = False

    def _init_model(self, parameters, **kwargs):
        self.constants.num_quantum_states = 2
        self.constants.num_classical_coordinates = 1
        self.constants.classical_coordinate_mass = np.array(
            [self.constants.get("mass", self.default_constants.get("mass"))]
        )
        self.constants.classical_coordinate_weight = np.array([1.0])
        self.constants.init_position = np.array([self.constants.get("init_position")])
        self.constants.init_momentum = np.array([self.constants.get("init_momentum")])
        return

    def h_qc(self, parameters, **kwargs):
        """
        Quantum-classical Hamiltonian for Tully's third problem.
        """
        z = kwargs["z"]
        batch_size = len(z)
        num_quantum_states = self.constants.num_quantum_states
        A = self.constants.get("A")
        B = self.constants.get("B")
        C = self.constants.get("C")
        # Calculate q.
        m = self.constants.classical_coordinate_mass[np.newaxis, :]
        h = self.constants.classical_coordinate_weight[np.newaxis, :]
        q = functions.z_to_q(z, m, h)[:, 0]
        # Calculate matrix elements.
        v_12 = np.zeros(batch_size, dtype=complex)
        v_12[q >= 0] = B * (2.0 - np.exp(-C * q))[q >= 0]
        v_12[q < 0] = B * np.exp(C * q)[q < 0]
        v_11 = np.ones(batch_size) * A
        # Assemble Hamiltonian.
        h_qc = np.zeros(
            (batch_size, num_quantum_states, num_quantum_states), dtype=complex
        )
        h_qc[:, 0, 0] = v_11
        h_qc[:, 0, 1] = v_12
        h_qc[:, 1, 0] = v_12
        h_qc[:, 1, 1] = -v_11
        return h_qc

    def dh_qc_dzc(self, parameters, **kwargs):
        """
        Gradient w.r.t. to the conjugate z coordinate of the quantum-classical Hamiltonian
        for Tully's third problem.
        """
        z = kwargs["z"]
        batch_size = len(z)
        num_quantum_states = self.constants.num_quantum_states
        num_classical_coordinates = self.constants.num_classical_coordinates
        B = self.constants.get("B")
        C = self.constants.get("C")
        # Calculate q.
        m = self.constants.classical_coordinate_mass[np.newaxis, :]
        h = self.constants.classical_coordinate_weight[np.newaxis, :]
        q = functions.z_to_q(z, m, h)[:, 0]
        # Calculate phase-space gradients.
        dv_12_dq = np.zeros(batch_size, dtype=complex)
        dv_12_dq[q >= 0.0] = B * C * np.exp(-C * q)[q >= 0.0]
        dv_12_dq[q < 0.0] = B * C * np.exp(C * q)[q < 0.0]
        # Convert to complex gradients.
        dv_12_dzc = functions.dqdp_to_dzc(dv_12_dq, None, m[0], h[0])
        # Assemble indices.
        batch_idx = np.repeat(np.arange(batch_size, dtype=int), 2)
        coord_idx = np.zeros(2 * batch_size, dtype=int)
        state_i_idx = np.tile(np.array([0, 1], dtype=int), batch_size)
        state_j_idx = np.tile(np.array([1, 0], dtype=int), batch_size)
        inds = (batch_idx, coord_idx, state_i_idx, state_j_idx)
        # Assemble matrix elements.
        mels = np.empty(2 * batch_size, dtype=complex)
        mels[0::2] = dv_12_dzc
        mels[1::2] = dv_12_dzc
        # Assemble shape.
        shape = (
            batch_size,
            num_classical_coordinates,
            num_quantum_states,
            num_quantum_states,
        )
        return inds, mels, shape

    ingredients = [
        ("h_q", ingredients.h_q_two_level),
        ("h_qc", h_qc),
        ("h_c", ingredients.h_c_free),
        ("dh_qc_dzc", dh_qc_dzc),
        ("dh_c_dzc", ingredients.dh_c_dzc_free),
        ("init_classical", ingredients.init_classical_definite_position_momentum),
        ("hop", ingredients.hop_free),
        ("_init_model", _init_model),
    ]



"""
This module contains the Simulation class.
"""

import logging
import numpy as np
from qclab.constants import Constants

logger = logging.getLogger(__name__)


class Simulation:
    """
    Simulation class for holding simulation components.
    """

    def __init__(self, settings=None):
        if settings is None:
            settings = {}
        self.default_settings = {
            "tmax": 10.0,
            "dt_update": 0.001,
            "dt_collect": 0.1,
            "num_trajs": 100,
            "batch_size": 25,
            "progress_bar": True,
            "debug": False,
        }
        # Merge default settings with user-provided settings.
        settings = {**self.default_settings, **settings}
        # Construct a Constants object to hold settings.
        self.settings = Constants()
        # Put settings from the dictionary into the Constants object.
        for key, val in settings.items():
            setattr(self.settings, key, val)
        # Set the initial algorithm and model to None.
        self.algorithm = None
        self.model = None
        # Initialize a dictionary to hold the initial state.
        self.initial_state = {}

    def initialize_timesteps(self):
        """
        Initialize the timesteps for the simulation based on the parameters.

        Adjusts ``dt_collect`` to be smallest integer multiple of ``dt_update`` that is nonzero.
        Then adjusts ``tmax`` to be the closest integer multiple of ``dt_collect``.
        """
        tmax = self.settings.get("tmax")
        dt_update = self.settings.get("dt_update")
        dt_collect = self.settings.get("dt_collect")

        logger.info(
            "Initializing timesteps with tmax=%s, dt_update=%s, dt_collect=%s",
            tmax,
            dt_update,
            dt_collect,
        )

        # dt_collect_n is the number of update timesteps that defines a collect
        # timestep.
        dt_collect_n = np.round(dt_collect / dt_update).astype(int)
        if dt_collect_n == 0:
            dt_collect_n = 1
            dt_collect = dt_update
            logger.warning(
                "dt_update is greater than dt_collect, setting dt_collect to dt_update."
            )

        # tmax_n is the number of update timesteps that defines the total
        # simulation time.
        self.settings.tmax_n = np.round(tmax / dt_collect).astype(int) * dt_collect_n
        self.settings.tmax = self.settings.tmax_n * dt_update
        self.settings.dt_collect_n = dt_collect_n
        self.settings.dt_collect = dt_collect_n * dt_update
        # t_update is the update time array for the simulation.
        self.settings.t_update = np.arange(0, self.settings.tmax_n + 1, 1) * dt_update
        # t_update_n is the update time array in terms of the number of update steps.
        self.settings.t_update_n = np.arange(0, self.settings.tmax_n + 1, 1)
        # t_collect_n is the collect time array for the simulation in terms of
        # the number of update steps.
        self.settings.t_collect_n = np.arange(
            0,
            self.settings.tmax_n + self.settings.dt_collect_n,
            self.settings.dt_collect_n,
        )
        # t_collect is the collect time array for the simulation.
        self.settings.t_collect = self.settings.t_collect_n * dt_update

        logger.info(
            "Initialization finished with tmax=%s, dt_update=%s, dt_collect=%s",
            self.settings.tmax,
            self.settings.get("dt_update"),
            self.settings.dt_collect,
        )



"""
This module contains the Model class.
"""

import copy
from qclab.constants import Constants


class Model:
    """
    Model class for defining model constants and ingredients.
    """

    def __init__(self, default_constants=None, constants=None):
        if constants is None:
            constants = {}
        if default_constants is None:
            default_constants = {}
        # Merge default constants with user-provided constants.
        constants = {**default_constants, **constants}
        # Construct a Constants object to hold constants.
        # Provide an initialization function to reinitialize
        # internal model constants upon changing any constant.
        self.constants = Constants(self.initialize_constants)
        # Put constants from the dictionary into the Constants object.
        for key, val in constants.items():
            setattr(self.constants, key, val)
        # Mark the constants as initialized.
        self.constants._init_complete = True
        # Copy the ingredients to ensure they are not shared across instances.
        self.ingredients = copy.deepcopy(self.ingredients)
        # Flags to indicate if the quantum Hamiltonian and quantum-classical
        # gradients need to be updated.
        self.update_h_q = True
        self.update_dh_qc_dzc = True
        self.initialize_constants()

    def get(self, ingredient_name):
        """
        Retrieve an ingredient by name.
        If the ingredient is not found or is None, returns (None, False).
        If the ingredient is found and not None, returns (ingredient, True).

        Args
        -----------
        ingredient_name : str
            Name of the ingredient to search for.

        Returns
        -----------
        tuple[callable | None, bool]: The ingredient function (or None if
            not found) and a flag indicating whether it exists.
        """
        for ingredient in self.ingredients[::-1]:
            if ingredient[0] == ingredient_name and not (ingredient[1] is None):
                return ingredient[1], True
            if ingredient[0] == ingredient_name and (ingredient[1] is None):
                return None, False
        return None, False

    def initialize_constants(self):
        """
        Initialize the constants for the model and ingredients.
        Calls ingredients with names starting with "_init_".
        """
        for ingredient in self.ingredients[::-1]:
            if ingredient[0].startswith("_init_") and ingredient[1] is not None:
                ingredient[1](self, None)
        return

    ingredients = []



"""
This module contains the Algorithm class.
"""

import copy
from qclab.constants import Constants


class Algorithm:
    """
    Algorithm class for defining and executing algorithm recipes.
    """

    def __init__(self, default_settings=None, settings=None):
        if settings is None:
            settings = {}
        if default_settings is None:
            default_settings = {}
        # Merge default settings with user-provided settings.
        settings = {**default_settings, **settings}
        # Construct a Constants object to hold settings.
        self.settings = Constants()
        # Put settings from the dictionary into the Constants object.
        for key, val in settings.items():
            setattr(self.settings, key, val)
        # Copy the recipes and output variables to ensure they are not shared
        # across instances.
        self.initialization_recipe = copy.deepcopy(self.initialization_recipe)
        self.update_recipe = copy.deepcopy(self.update_recipe)
        self.collect_recipe = copy.deepcopy(self.collect_recipe)

    initialization_recipe = []
    update_recipe = []
    collect_recipe = []

    def execute_recipe(self, sim, state, parameters, recipe):
        """
        Carry out the given recipe for the simulation by running
        each task in the recipe.

        .. rubric:: Args
        sim: Simulation
            The Simulation object containing settings and parameters.
        state: dict
            The dictionary containing the current simulation state.
        parameters: dict
            The dictionary containing the current simulation parameters.
        recipe: list
            The list of functions to execute in order.

        .. rubric:: Returns
        state: dict
            The updated simulation state after executing the recipe.
        parameters: dict
            The updated simulation parameters after executing the recipe.
        """
        for func in recipe:
            state, parameters = func(sim, state, parameters)
        return state, parameters
